#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with spanning-tree in EOS

This module provides an API for working with spanning-tree configuration
in EOS.  This includes both global spanning-tree configuration as well as
interface config.

Global Parameters:
    mode (string): The spanning-tree operational mode.  Accepted values
        for this version are 'mstp' or 'none'.  This configuration
        parameter is not defaultable

    interfaces (StpInterfaces): The collection of STP enabled interfaces.

Interface Parameters:
    name (string): The name of the interface the STP configuration is in
        reference to.  The interface name is the full interface identifier

    portfast (string): The portfast configuration value for the interface.
        Accepted values are 'edge', 'network', or 'disabled'

    bpduguard (boolean): True if the BPDU Guard feature is enabled on the
        interface or False if it is disabled
"""

import re

from pyeapi.api import Entity, EntityCollection


class Stp(Entity):
    """The Stp class implements global configuration for spanning-tree

    The spanning-tree protocol provides both global and interface
    configuration options.  This class is the top-level class that provides
    access to all spanning-tree configuration options supported.

    Example:
        The below example demonstrates how to use the STP class to work
        with both global configuration and interface configuration.

        >>> import pyeapi.resources.stp
        >>> stp = pyeapi.resources.stp.instance(node)
        >>> stp.set_mode('mstp')
        True
        >>> stp.interfaces.set_bpduguard('Ethernet1', True)
        True

    Attributes:
        interfaces (StpInterfaces): An instance for configuration spanning-tree
            configuration interfaces

        instances (StpInstances): An instance object for working with
            spanning-tree global instances

    """

    def __init__(self, *args, **kwargs):
        super(Stp, self).__init__(*args, **kwargs)
        self._interfaces = None
        self._instances = None

    def get(self):
        """Returns the spanning-tree configuration as a dict object

        The dictionary object represents the entire spanning-tree
        configuration derived from the nodes running config.  This
        includes both globally configuration attributes as well as
        interfaces and instances.  See the StpInterfaces and StpInstances
        classes for the key/value pair definitions.

        Note:
            See the individual classes for detailed message structures

        Returns:
            A Python dictionary object of key/value pairs the represent
            the entire supported spanning-tree configuration::

                {
                    "mode": [mstp, none],
                    "interfaces": {...},
                    "instances": {...}
                }
        """
        return dict(interfaces=self.interfaces.getall(),
                    instances=self.instances.getall())

    @property
    def interfaces(self):
        if self._interfaces is not None:
            return self._interfaces
        self._interfaces = StpInterfaces(self.node)
        return self._interfaces

    @property
    def instances(self):
        if self._instances is not None:
            return self._instances
        self._instances = StpInstances(self.node)
        return self._instances

    def set_mode(self, value=None, default=False, disable=False):
        """Configures the global spanning-tree mode

        Note:
            This configuration parameter is not defaultable

        Args:
            value (string): The value to configure the global spanning-tree
                mode of operation.  Valid values include 'mstp', 'none'
            default (bool): Set the global spanning-tree mode to default.
            disable (bool): Negate the global spanning-tree mode.

        Returns:
            True if the configuration operation succeeds otherwise False

        Raises:
            ValueError if the value is not in the accepted range
        """
        if not default and not disable:
            if value not in ['mstp', 'none']:
                raise ValueError("Specified value must be one of "
                                 "'mstp', 'none'")
        cmds = self.command_builder('spanning-tree mode', value=value,
                                    default=default, disable=disable)
        return self.configure(cmds)


class StpInstances(EntityCollection):
    """Provides a configuration resource for spanning-tree instances

    This class provides an API for working with spanning-tree instances from
    the global configuration.  Spanning tree instances work with MST
    configuration
    """

    def getall(self):
        # TODO: (privateip, 20150106) stubbed out, needs implementation
        return dict()


class StpInterfaces(EntityCollection):
    """Provides a configuration resource for spanning-tree interfaces

    This class provides an API for working with spanning-tree interface
    configurations.  It provides access to managing specific interface
    spanning-tree configuration options.  Note that spanning-tree interfaces
    cannot be created or deleted.
    """

    def get(self, name):
        """Returns the specified interfaces STP configuration resource

        The STP interface resource contains the following

            * name (str): The interface name
            * portfast (bool): The spanning-tree portfast admin state
            * bpduguard (bool): The spanning-tree bpduguard admin state
            * portfast_type (str): The spanning-tree portfast <type> value.
                Valid values include "edge", "network", "normal"

        Args:
            name (string): The interface identifier to retrieve the config
                for.  Note: Spanning-tree interfaces are only supported on
                Ethernet and Port-Channel interfaces

        Returns:
            dict: A resource dict object that represents the interface
                configuration.

            None: If the specified interace is not a STP port
        """
        if not isvalidinterface(name):
            return None

        config = self.get_block(r'^interface\s%s$' % name)
        resp = dict()
        resp.update(self._parse_bpduguard(config))
        resp.update(self._parse_portfast(config))
        resp.update(self._parse_portfast_type(config))
        return resp

    def _parse_bpduguard(self, config):
        value = 'spanning-tree bpduguard enable' in config
        return dict(bpduguard=value)

    def _parse_portfast(self, config):
        value = 'no spanning-tree portfast' not in config
        return dict(portfast=value)

    def _parse_portfast_type(self, config):
        if 'spanning-tree portfast network' in config:
            value = 'network'
        elif 'no spanning-tree portfast' in config:
            value = 'normal'
        else:
            value = 'edge'
        return dict(portfast_type=value)

    def getall(self):
        """Returns the collection of STP interfaces

        This method will return all of the configured spanning-tree
        interfaces from the current nodes configuration.

        Returns:
            A Python dictionary object that represents all configured
                spanning-tree interfaces indexed by interface name.
        """
        interfaces_re = re.compile(r'(?<=^interface\s)(.+)$', re.M)

        response = dict()
        for name in interfaces_re.findall(self.config):
            if name[0:2] in ['Et', 'Po']:
                interface = self.get(name)
                if interface:
                    response[name] = interface
        return response

    def configure_interface(self, name, cmds):
        if not isvalidinterface(name):
            raise ValueError('invalid interface value specified')
        return super(StpInterfaces, self).configure_interface(name, cmds)

    def set_portfast_type(self, name, value='normal'):
        """Configures the portfast value for the specified interface

        Args:
            name (string): The interface identifier to configure.  The name
                must be the full interface name (eg Ethernet1, not Et1).

            value (string): The value to configure the portfast setting to.
                Valid values include "edge", "network", "normal".  The
                default value is "normal"

        Returns:
            True if the command succeeds, otherwise False

        Raises:
            ValueError: Raised if an invalid interface name or value is
                specified

        """
        if value not in ['network', 'edge', 'normal', None]:
            raise ValueError('invalid portfast type value specified')

        cmds = ['spanning-tree portfast %s' % value]
        if value == 'edge':
            cmds.append('spanning-tree portfast auto')
        return self.configure_interface(name, cmds)

    def set_portfast(self, name, value=None, default=False, disable=False):
        """Configures the portfast value for the specified interface

        Args:
            name (string): The interface identifier to configure.  The name
                must be the full interface name (eg Ethernet1, not Et1)

            value (bool): True if portfast is enabled otherwise False

            default (bool): Configures the portfast parameter to its default
                value using the EOS CLI default config command

            disable (bool): Negates the portfast parameter using the EOS
                CLI no config command

        Returns:
            True if the command succeeds, otherwise False

        Raises:
            ValueError: Rasied if an invalid interface name is specified

            TypeError: Raised if the value keyword argument does not evaluate
                to a valid boolean

        """
        if value is False:
            disable = True
        string = 'spanning-tree portfast'
        cmds = self.command_builder(string, value=value, default=default,
                                    disable=disable)
        return self.configure_interface(name, cmds)

    def set_bpduguard(self, name, value=False, default=False, disable=False):
        """Configures the bpduguard value for the specified interface

        Args:
            name (string): The interface identifier to configure.  The name
                must be the full interface name (eg Ethernet1, not Et1)

            value (bool): True if bpduguard is enabled otherwise False

            default (bool): Configures the bpduguard parameter to its default
                value using the EOS CLI default config command

            disable (bool): Negates the bpduguard parameter using the EOS
                CLI no config command

        Returns:
            True if the command succeeds, otherwise False

        Raises:
            ValueError: Rasied if an invalid interface name is specified

            TypeError: Raised if the value keyword argument does not evaluate
                to a valid boolean

        """
        value = 'enable' if value else 'disable'
        string = 'spanning-tree bpduguard'
        cmds = self.command_builder(string, value=value, default=default,
                                    disable=disable)
        return self.configure_interface(name, cmds)


def isvalidinterface(value):
    """Checks value to see if it could be a spanning-tree interface

    This function will check the value and return a boolean whether or not
    the interface could be a spanning-tree interface

    Note:
        This function only checks *if* the interface could be a spanning-tree
            interface but does not check if it *is* configured as a
            spanning-tree interface

    Args:
        value (string): The interface name to validate

    Returns:
        True if it could be a spanning-tree interface, otherwise False
    """
    return value[0:2] in ['Et', 'Po']


def instance(api):
    return Stp(api)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Python Client for eAPI

This module provides the client for eAPI.  It provides the primary functions
for building applications that work with Arista EOS eAPI-enabled nodes.  The
first function is to provide a client for sending and receiving eAPI
request and response objects on a per node basis.  The second function
provides a library for building API enabled data models for configuring
EOS nodes.

This library allows for creating connections to EOS eAPI enabled nodes using
the connect or connect_to function.  Both functions will return an instance
of a Node object that can be used to send and receive eAPI commands.  The
Node object can autoload API modules for a structured object oriented
approach to configuring the EOS node with native Python objects.

Example:

    >>> import pyeapi
    >>> conn = pyeapi.connect(host='10.1.1.1', transport='http')
    >>> conn.execute(['show version'])
    {u'jsonrpc': u'2.0', u'result': [{u'memTotal': 2028008, u'version':
    u'4.14.5F', u'internalVersion': u'4.14.5F-2209869.4145F', u'serialNumber':
    u'', u'systemMacAddress': u'00:0c:29:f5:d2:7d', u'bootupTimestamp':
    1421765066.11, u'memFree': 213212, u'modelName': u'vEOS', u'architecture':
    u'i386', u'internalBuildId': u'f590eed4-1e66-43c6-8943-cee0390fbafe',
    u'hardwareRevision': u''}], u'id': u'4312565648'}

    >>> node = pyeapi.connect_to('veos01')
    >>> node.enable('show version')
    {u'jsonrpc': u'2.0', u'result': [{u'memTotal': 2028008, u'version':
    u'4.14.5F', u'internalVersion': u'4.14.5F-2209869.4145F', u'serialNumber':
    u'', u'systemMacAddress': u'00:0c:29:f5:d2:7d', u'bootupTimestamp':
    1421765066.11, u'memFree': 213212, u'modelName': u'vEOS', u'architecture':
    u'i386', u'internalBuildId': u'f590eed4-1e66-43c6-8943-cee0390fbafe',
    u'hardwareRevision': u''}], u'id': u'4312565648'}

Additionally the node object can automatically load API modules to work
with the resources in the configuration.  The API autoloader supports
automatic loading of modules in pyeapi.api as well as provides the ability
to build custom API modules to be loaded from a different namespace.

Example:

    >>> import pyeapi
    >>> node = pyeapi.connect_to('veos01')
    >>> node.api('vlans').get(1)
    {'state': 'active', 'name': 'default', 'vlan_id': 1, 'trunk_groups': []}

The API autoloader loads API modules by their filename.

The following objects are provide in this module for creating clients to
interface with eAPI.

Node -- Creates an instance of a node object that represents a single EOS
device.  Each EOS device to be managed should have a Node instance

Config -- A subclass of ConfigParser.SafeConfigParser that handles the
configuration file.  The configuration file is an INI style file that
contains the settings for nodes used by the connect_to function.

"""
from uuid import uuid4
import os
import re

from functools import lru_cache

# Note: SafeConfigParser is deprecated and replaced by ConfigParser
from configparser import ConfigParser as SafeConfigParser
from configparser import Error as SafeConfigParserError

from pyeapi.utils import load_module, make_iterable, debug, CliVariants

from pyeapi.eapilib import HttpEapiConnection, HttpsEapiConnection
from pyeapi.eapilib import HttpsEapiCertConnection
from pyeapi.eapilib import HttpEapiSessionConnection
from pyeapi.eapilib import HttpsEapiSessionConnection
from pyeapi.eapilib import SocketEapiConnection, HttpLocalEapiConnection
from pyeapi.eapilib import CommandError

CONFIG_SEARCH_PATH = ['~/.eapi.conf', '/mnt/flash/eapi.conf']

TRANSPORTS = {
    'socket': SocketEapiConnection,
    'http_local': HttpLocalEapiConnection,
    'http': HttpEapiConnection,
    'http_session': HttpEapiSessionConnection,
    'https': HttpsEapiConnection,
    'https_certs': HttpsEapiCertConnection,
    'https_session': HttpsEapiSessionConnection,
}

DEFAULT_TRANSPORT = 'https'


class Config(SafeConfigParser):
    """Conifguration instance for managing the eapi.conf file.

    This class provides an instance for handling the configuration file.  It
    should normally need to be instantiated.  A single config object is
    instantiated by the module for working with the config.

    Attributes:
        filename (str): The full path to the loaded filename

    Args:
        filename(str): The full path to the filename to be loaded when the
            object is instantiated.

    """
    def __init__(self, filename=None):
        SafeConfigParser.__init__(self)

        self.filename = filename
        self.tags = dict()

        self.autoload()

    @property
    def connections(self):
        """
        Returns all of the loaded connections names as a list
        """
        conn = lambda x: str(x).replace('connection:', '')
        return [conn(name) for name in self.sections()]

    def autoload(self):
        """ Loads the eapi.conf file

        This method will use the module variable CONFIG_SEARCH_PATH to
        attempt to locate a valid eapi.conf file if a filename is not already
        configured.   This method will load the first eapi.conf file it
        finds and then return.

        The CONFIG_SEARCH_PATH can be overridden using an environment variable
        by setting EAPI_CONF.

        """
        path = list(CONFIG_SEARCH_PATH)
        if 'EAPI_CONF' in os.environ:
            path = os.environ['EAPI_CONF']
        elif self.filename:
            path = self.filename

        path = make_iterable(path)

        for filename in path:
            filename = os.path.expanduser(filename)
            if os.path.exists(filename):
                self.filename = filename
                return self.read(filename)

        self._add_default_connection()

    def read(self, filename):
        """Reads the file specified by filename

        This method will load the eapi.conf file specified by filename into
        the instance object.  It will also add the default connection localhost
        if it was not defined in the eapi.conf file

        Args:
            filename (str): The full path to the file to load
        """

        try:
            SafeConfigParser.read(self, filename)
        except SafeConfigParserError as exc:
            # Ignore file and syslog a message on SafeConfigParser errors
            msg = ("%s: parsing error in eapi conf file: %s" %
                   (type(exc).__name__, filename))
            debug(msg)

        self._add_default_connection()

        for name in self.sections():
            if name.startswith('connection:') and \
               'host' not in dict(self.items(name)):

                self.set(name, 'host', name.split(':')[1])
        self.generate_tags()

    def _add_default_connection(self):
        """Checks the loaded config and adds the localhost profile if needed

        This method wil load the connection:localhost profile into the client
        configuration if it is not already present.
        """
        if not self.get_connection('localhost'):
            self.add_connection('localhost', transport='socket')

    def generate_tags(self):
        """ Generates the tags with collection with hosts
        """
        self.tags = dict()
        for section in self.sections():
            if self.has_option(section, 'tags'):
                tags = self.get(section, 'tags')
                for tag in [str(t).strip() for t in tags.split(',')]:
                    if tag not in self.tags:
                        self.tags[tag] = list()
                    self.tags[tag].append(section.split(':')[1])

    def load(self, filename):
        """Loads the file specified by filename

        This method works in conjunction with the autoload method to load the
        file specified by filename.

        Args:
            filename (str): The full path to the file to be loaded
        """
        self.filename = filename
        self.reload()

    def reload(self):
        """Reloades the configuration

        This method will reload the configuration instance using the last
        known filename.  Note this method will initially clear the
        configuration and reload all entries.

        """
        for section in self.sections():
            self.remove_section(section)
        self.autoload()

    def get_connection(self, name):
        """Returns the properties for a connection name

        This method will return the settings for the configuration specified
        by name.  Note that the name argument should only be the name.

        For instance, give the following eapi.conf file

        .. code-block:: ini

            [connection:veos01]
            transport: http

        The name to use to retrieve the configuration would be veos01

            >>> pyeapi.client.config.get_connection('veos01')

        Args:
            name (str): The name of the connection to return

        Returns:
            A Python dictionary object of key/value pairs that represent
                the node configuration.  If the name provided in the argument
                is not found, then None is returned.
        """
        name = 'connection:{}'.format(name)
        if not self.has_section(name):
            return None
        return dict(self.items(name))

    def add_connection(self, name, **kwargs):
        """Adds a connection to the configuration

        This method will add a connection to the configuration.  The connection
        added is only available for the lifetime of the object and is not
        persisted.

        Note:
            If a call is made to load() or reload(), any connections added
            with this method must be re-added to the config instance

        Args:
            name (str): The name of the connection to add to the config.  The
                name provided will automatically be prepended with the string
                connection:
            **kwargs (dict); The set of properties used to provide the node
                configuration

        """
        name = 'connection:{}'.format(name)
        self.add_section(name)
        for key, value in list(kwargs.items()):
            self.set(name, key, value)
        self.generate_tags()

# TODO: This is a global variable (in the module) - to review the impact on
# having a shared state for the config file.


config = Config()


def load_config(filename):
    """Function method that loads a conf file

    This function will load the file specified by filename into the config
    instance.   Its a convenience function that calls load on the config
    instance

    Args:
        filename (str): The full path to the filename to load
    """
    return config.load(filename)


def config_for(name):
    """ Function to get settings for named config

    This function will return the settings for a specific connection as
    specified by name.  Its a convenience function that calls get_connection
    on the global config instance

    Args:
        name (str): The name of the connection to return.  The connection
            name is specified as the string right of the : in the INI file

    Returns:
        A Python dictionary object of key/value pairs that represent the
            nodes configuration settings from the config instance
    """
    return config.get_connection(name)


def hosts_for_tag(tag):
    """ Returns the hosts assocated with the specified tag

    This function will return the hosts assoicated with the tag specified
    in the argument.  It will return an array of connecition names.

    Args:
        tag (str): The name of the tag to retrieve the list of hosts for

    Returns:
        list: A Python list object that includes the list of hosts assoicated
            with the specified tag.

        None: If the specified tag does not exist, then None is returned.
    """
    return config.tags.get(tag)


def make_connection(transport, **kwargs):
    """ Creates a connection instance based on the transport

    This function creates the EapiConnection object based on the desired
    transport.  It looks up the transport class in the TRANSPORTS global
    dictionary.

    Args:
        transport (string): The transport to use to create the instance.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        An instance of a connection object based on the transport

    Raises:
        TypeError: A TypeError is raised if the transport keyword is not
            found in the list (keys) of available transports.
    """
    if transport not in TRANSPORTS:
        raise TypeError('invalid transport specified')
    klass = TRANSPORTS[transport]
    return klass(**kwargs)


def connect(transport=None, host='localhost', username='admin',
            password='', port=None, key_file=None, cert_file=None,
            ca_file=None, timeout=60, return_node=False, context=None,
            **kwargs):
    """ Creates a connection using the supplied settings

    This function will create a connection to an Arista EOS node using
    the arguments.  All arguments are optional with default values.

    Args:
        transport (str): Specifies the type of connection transport to use.
            Valid values for the connection are socket, http_local, http,
            https, https_certs, http_session, and https_session.  The default
            value is specified in DEFAULT_TRANSPORT
        host (str): The IP addres or DNS host name of the connection device.
            The default value is 'localhost'
        username (str): The username to pass to the device to authenticate
            the eAPI connection.   The default value is 'admin'
        password (str): The password to pass to the device to authenticate
            the eAPI connection.  The default value is ''
        port (int): The TCP port of the endpoint for the eAPI connection.  If
            this keyword is not specified, the default value is automatically
            determined by the transport type. (http=80, https=443)
        key_file (str): Path to private key file for ssl validation
        cert_file (str): Path to PEM formatted cert file for ssl validation
        ca_file (str): Path to CA PEM formatted cert file for ssl validation
        timeout (int): timeout
        context (ssl.SSLContext): ssl object's context. The default is None
        return_node (bool): Returns a Node object if True, otherwise
            returns an EapiConnection object.


    Returns:
        An instance of an EapiConnection object for the specified transport.

    Note:
        Python 3.10 increases security strength of the TLS stack by among other
        things using a stronger (than 3.9) default cipher suite. Thus programs
        relying on the https transport and using the default cypher suite that
        used to work in prior python versions may fail with the error:
        ``[SSL: SSLV3_ALERT_HANDSHAKE_FAILURE] sslv3 alert handshake failure``.
        The solution to that issue is to configure the https web server to use
        a stronger cipher suite.

        If the solution is not attainable, then a work-around might be
        considered (weighing all due implications) - one could pass an ssl
        context where cipher can be specified::

          import pyeapi
          import ssl
          ...
          ctx = ssl.create_default_context()
          ctx.set_ciphers('DEFAULT')          # set a preferred cipher
          ctx.check_hostname = False          # for the sake of example
          ctx.verify_mode = ssl.CERT_NONE     # do it w/o certificate
          ...
          cc = pyeapi.client.connect( host=host_name, context=ctx )
    """
    transport = transport or DEFAULT_TRANSPORT
    connection = make_connection(transport, host=host, username=username,
                                 password=password, key_file=key_file,
                                 cert_file=cert_file, ca_file=ca_file,
                                 port=port, timeout=timeout, context=context)
    if return_node:
        return Node(connection, transport=transport, host=host,
                    username=username, password=password, key_file=key_file,
                    cert_file=cert_file, ca_file=ca_file, port=port, **kwargs)
    return connection


class Node(object):
    """Represents a single device for sending and receiving eAPI messages

    The Node object provides an instance for communicating with Arista EOS
    devices.  The Node object provides easy to use methods for sending both
    enable and config commands to the device using a specific transport.  This
    object forms the base for communicating with devices.

    Attributes:
        connection (EapiConnection): The connection property represents the
            underlying transport used by the Node object to communicate
            with the device using eAPI.
        running_config (str): The running-config from the device.  This
            property is lazily loaded and refreshed over the life cycle of
            the instance.
        startup_config (str): The startup-config from the device.  This
            property is lazily loaded and refreshed over the life cycle of
            the instance.
        autorefresh (bool): If True, the running-config and startup-config are
            refreshed on config events.  If False, then the config properties
            must be manually refreshed.
        config_defaults (bool): If True, the default config options will be
            shown in the running-config output
        settings (dict): Provides access to the settings used to create the
            Node instance.

    Args:
        connection (EapiConnection): An instance of EapiConnection used as the
            transport for sending and receiving eAPI requests and responses.
        **kwargs: An arbitrary list of keyword arguments
    """
    def __init__(self, connection, **kwargs):
        self._connection = connection
        self._running_config = None
        self._startup_config = None
        self._version = None
        self._version_number = None
        self._model = None
        self._session_name = None

        self._enablepwd = kwargs.get('enablepwd')
        self.autorefresh = kwargs.get('autorefresh', True)
        self.config_defaults = kwargs.get('config_defaults', True)
        self.settings = kwargs

    def __str__(self):
        return 'Node(connection=%s)' % str(self._connection)

    def __repr__(self):
        return 'Node(connection=%s)' % repr(self._connection)

    @property
    def connection(self):
        return self._connection

    @property
    def running_config(self):
        if self._running_config is not None:
            return self._running_config
        params = 'all' if self.config_defaults else None
        self._running_config = self.get_config(params=params,
                                               as_string=True)
        return self._running_config

    @property
    def startup_config(self):
        if self._startup_config is not None:
            return self._startup_config
        self._startup_config = self.get_config('startup-config',
                                               as_string=True)
        return self._startup_config

    @property
    def version(self):
        if self._version:
            return self._version
        self._get_version_properties()
        return self._version

    @property
    def version_number(self):
        if self._version_number:
            return self._version_number
        self._get_version_properties()
        return self._version_number

    @property
    def model(self):
        if self._model:
            return self._model
        self._get_version_properties()
        return self._model

    def _get_version_properties(self):
        """Parses version and model information out of 'show version' output
        and uses the output to populate class properties.
        """
        # Parse out version info
        output = self.enable('show version')
        self._version = str(output[0]['result']['version'])
        match = re.match(r'[\d.\d]+', str(output[0]['result']['version']))
        if match:
            self._version_number = str(match.group(0))
        else:
            self._version_number = str(output[0]['result']['version'])
        # Parse out model number
        match = re.search(r'\d\d\d\d', str(output[0]['result']['modelName']))
        if match:
            self._model = str(match.group(0))
        else:
            self._model = str(output[0]['result']['modelName'])

    def enable_authentication(self, password):
        """Configures the enable mode authentication password

        EOS supports an additional password authentication mechanism for
        sessions that want to switch to executive (or enable) mode.  This
        method will configure the password, if required, for entering
        executive mode

        Args:
            password (str): The password string in clear text used to
                authenticate to exec mode
        """
        self._enablepwd = str(password).strip()

    def config(self, commands, **kwargs):
        """Configures the node with the specified commands

        This method is used to send configuration commands to the node.  It
        will take either a string, list or CliVariants type and prepend the
        necessary commands to put the session into config mode.
        pyeapi.utils.CliVariants facilitates alternative executions to commands
        sequence until one variant succeeds or all fail

        Args:
            commands (str, list, CliVariants): The commands to send to the node
                in config mode. If the commands argument is an str or
                CliVariants type, it will be cast to a list.
                The list of commands will also be prepended with the necessary
                commands to put the session in config mode.
                CliVariants could be part of a list too, however only a single
                occurrence of CliVariants type in commands is supported.
                CliVariants type facilitates execution of alternative commands
                sequences, e.g.:
                ``config( [cli1, CliVariants( cli2, cli3 ), cli4] )``
                the example above can be translated into following sequence:
                ``config( [cli1, cli2, cli4] )``
                ``config( [cli1, cli3, cli4] )``
                CliVariants accepts 2 or more arguments of str, list type, or
                their mix. Each argument to CliVariants will be joined with the
                rest of commands and all command sequences will be tried until
                one variant succeeds. If all variants fail the last failure
                exception will be re-raised.

            **kwargs: Additional keyword arguments for expanded eAPI
                functionality. Only supported eAPI params are used in building
                the request

        Returns:
            The config method will return a list of dictionaries with the
                output from each command.  The function will strip the
                response from any commands it prepends.
        """
        def variant_cli_idx( cmds ):
            # return index of first occurrence of CliVariants type in cmds
            try:
                return [ type(v) for v in cmds ].index( CliVariants )
            except (ValueError):
                return -1

        cfg_call = self._configure_session if self._session_name \
            else self._configure_terminal

        if isinstance( commands, CliVariants ):
            commands = [ commands ]
        idx = variant_cli_idx( commands )
        if idx == -1:
            return cfg_call( commands, **kwargs )

        # commands contain CliVariants obj, e.g.: [ '...', CliVariants, ... ]
        err = None
        for variant in commands[ idx ].variants:
            cmd = commands[ :idx ] + variant + commands[ idx + 1: ]
            try:
                return cfg_call( cmd, **kwargs )
            except (CommandError) as exp:
                err = exp
        raise err  # re-raising last occurred CommandError


    def _configure_terminal(self, commands, **kwargs):
        """Configures the node with the specified commands with leading
        "configure terminal"
        """
        commands = make_iterable(commands)
        commands = list(commands)

        # push the configure command onto the command stack
        commands.insert(0, 'configure terminal')
        response = self.run_commands(commands, **kwargs)
        # after config change the _chunkify lru_cache has to be cleared
        self._chunkify.cache_clear()

        if self.autorefresh:
            self.refresh()

        # pop the configure command output off the stack
        response.pop(0)

        return response

    def _configure_session(self, commands, **kwargs):
        """Configures the node with the specified commands with leading
        "configure session <session name>"
        """
        if not self._session_name:
            raise CommandError(-1, 'Not currently in a session')

        commands = make_iterable(commands)
        commands = list(commands)

        # push the configure command onto the command stack
        commands.insert(0, 'configure session %s' % self._session_name)
        response = self.run_commands(commands, **kwargs)
        # after config change the _chunkify lru_cache has to be cleared
        self._chunkify.cache_clear()

        # pop the configure command output off the stack
        response.pop(0)

        return response

    @lru_cache(maxsize=None)
    def _chunkify( self, config, indent=0 ):
        """parse device config and return a dict holding sections and
        sub-sections:
        - a section always begins with a line with zero indents,
        - a sub-section always begins with an indented line
        a (sub)section typically contains a begin line (with a lower indent)
        and a body (with a higher indent). A section might be degenerative (no
        body, just the section line itself), while sub-sections always contain
        a sub-section line plus some body). E.g., here's a snippet of a section
        dict:
        { ...
          'spanning-tree mode none': 'spanning-tree mode none\n',
          ...
          'mac security': 'mac security\n  profile PR\n    cipher aes256-gcm',
          '   profile PR': '  profile PR\n    cipher aes256-gcm'
          ... }

        it's imperative that the most outer call is made with indent=0, as the
        indent parameter defines processing of nested sub-sections, i.e., if
        indent > 0, then it's a recursive call and `config` argument contains
        last parsed (sub)section, which in turn may contain sub-sections
        """
        def is_subsection_present( section, indent ):
            return any( line[ indent ] == ' ' for line in section )

        def get_indent( line ):
            return len( line ) - len( line.lstrip() )

        sections = {}
        key = None
        banner = None
        for line in config.splitlines( keepends=True )[ indent > 0: ]:
            line_rs = line.rstrip()
            if indent == 0:
                if banner:
                    sections[ banner ] += line
                    if line_rs == 'EOF':
                        banner = None
                    continue
                if line.startswith( 'banner ' ):
                    banner = line_rs
                    sections[ banner ] = line
                    continue
            if get_indent( line_rs ) > indent:  # i.e. subsection line
                # key is always expected to be set by now
                sections[ key ] += line
                continue
            subsection = sections.get( key, '' ).splitlines()[ 1: ]
            if subsection:
                sub_indent = get_indent( subsection[0] )
                if is_subsection_present( subsection, sub_indent ):
                    parsed = self._chunkify( sections[key], indent=sub_indent )
                    parsed.update( sections )
                    sections = parsed
            key = line_rs
            sections[ key ] = line
        return sections

    def section(self, regex, config='running_config'):
        """Returns a section of the config

        Args:
            regex (str): A valid regular expression used to select sections
                of configuration to return
            config (str): The configuration to return.  Valid values for config
                are "running_config" or "startup_config".  The default value
                is "running_config"

        Returns:
            The configuration section as a string object.
        """
        if config in ['running_config', 'startup_config']:
            config = getattr(self, config)
        chunked = self._chunkify(config)
        r = re.compile(regex)
        matching_keys = [k for k in chunked.keys() if r.search(k)]
        if len(matching_keys) == 0:
            raise TypeError('config section not found')
        matching_key = matching_keys[0]
        match = chunked[matching_key]
        return match

    def enable(self, commands, encoding='json', strict=False,
               send_enable=True, **kwargs):
        """Sends the array of commands to the node in enable mode

        This method will send the commands to the node and evaluate
        the results.  If a command fails due to an encoding error,
        then the command set will be re-issued individual with text
        encoding.

        Args:
            commands (list): The list of commands to send to the node

            encoding (str): The requested encoding of the command output.
                Valid values for encoding are JSON or text

            strict (bool): If False, this method will attempt to run a
                command with text encoding if JSON encoding fails
            send_enable (bool): If True the enable command will be
                               prepended to the command list automatically.
            **kwargs: Additional keyword arguments for expanded eAPI
                functionality. Only supported eAPI params are used in building
                the request

        Returns:
            A dict object that includes the response for each command along
                with the encoding

        Raises:
            TypeError:
                This method does not support sending configure
                commands and will raise a TypeError if configuration commands
                are found in the list of commands provided

                This method will also raise a TypeError if the specified
                encoding is not one of 'json' or 'text'

            CommandError: This method will raise a CommandError if any one
                of the commands fails.
        """
        commands = make_iterable(commands)

        if 'configure' in commands:
            raise TypeError('config mode commands not supported')

        results = list()
        # IMPORTANT: There are two keys (response, result) that both
        # return the same value. 'response' was originally placed
        # there in error and both are now present to avoid breaking
        # existing scripts. 'response' will be removed in a future release.
        if strict:
            responses = self.run_commands(commands, encoding, send_enable,
                                          **kwargs)
            for index, response in enumerate(responses):
                results.append(dict(command=commands[index],
                                    result=response,
                                    response=response,
                                    encoding=encoding))
        else:
            for command in commands:
                try:
                    resp = self.run_commands(command, encoding, send_enable,
                                             **kwargs)
                    results.append(dict(command=command,
                                        result=resp[0],
                                        encoding=encoding))
                except CommandError as exc:
                    if exc.error_code == 1003:
                        resp = self.run_commands(command, 'text', send_enable,
                                                 **kwargs)
                        results.append(dict(command=command,
                                            result=resp[0],
                                            encoding='text'))
                    else:
                        raise
        return results

    def run_commands(self, commands, encoding='json', send_enable=True,
                     **kwargs):
        """Sends the commands over the transport to the device

        This method sends the commands to the device using the nodes
        transport.  This is a lower layer function that shouldn't normally
        need to be used, preferring instead to use config() or enable().

        Args:
            commands (list): The ordered list of commands to send to the
                device using the transport
            encoding (str): The encoding method to use for the request and
                excpected response.
            send_enable (bool): If True the enable command will be
                               prepended to the command list automatically.
            **kwargs: Additional keyword arguments for expanded eAPI
                functionality. Only supported eAPI params are used in building
                the request

        Returns:
            This method will return the raw response from the connection
                which is a Python dictionary object.
        """
        commands = make_iterable(commands)

        # Some commands are multiline commands. These are banner commands and
        # SSL commands. So with this two lines we
        # can support those by passing commands by doing:
        # banner login MULTILINE: This is my banner.\nAnd I even support
        # multiple lines.
        # Why this? To be able to read a configuration from a file, split it
        # into lines and pass it as it is
        # to pyeapi without caring about multiline commands.
        commands = [{'cmd': c.split('MULTILINE:')[0],
                     'input': '%s\n' % (c.split('MULTILINE:')[1].strip())}
                    if 'MULTILINE:' in c else c for c in commands]

        if send_enable:
            if self._enablepwd:
                commands.insert(0, {'cmd': 'enable', 'input': self._enablepwd})
            else:
                commands.insert(0, 'enable')

        response = self._connection.execute(commands, encoding, **kwargs)

        # pop enable command from the response only if we sent enable
        if send_enable:
            response['result'].pop(0)

        return response['result']

    def api(self, name, namespace='pyeapi.api'):
        """Loads the specified api module

        This method is the API autoload mechanism that will load the API
        module specified by the name argument.  The API module will be loaded
        and look first for an initialize() function and secondly for an
        instance() function.  In both cases, the node object is passed to
        the module.

        Args:
            name (str): The name of the module to load.  The name should be
                the name of the python file to import
            namespace (str): The namespace to use to load the module.  The
                default value is 'pyeapi.api'

        Returns:
            The API module loaded with the node instance.
        """
        module = load_module('{}.{}'.format(namespace, name))
        if hasattr(module, 'initialize'):
            module.initialize(self)
        if hasattr(module, 'instance'):
            return module.instance(self)
        return module

    def get_config(self, config='running-config', params=None,
                   as_string=False):
        """ Retreives the config from the node

        This method will retrieve the config from the node as either a string
        or a list object.  The config to retrieve can be specified as either
        the startup-config or the running-config.

        Args:
            config (str): Specifies to return either the nodes startup-config
                or running-config.  The default value is the running-config
            params (str): A string of keywords to append to the command for
                retrieving the config.
            as_string (boo): Flag that determines the response.  If True, then
                the configuration is returned as a raw string.  If False, then
                the configuration is returned as a list.  The default value is
                False

        Returns:
            This method will return either a string or a list depending on the
            states of the as_string keyword argument.

        Raises:
            TypeError: If the specified config is not one of either
                'running-config' or 'startup-config'
        """
        if config not in ['startup-config', 'running-config']:
            raise TypeError('invalid config name specified')

        command = 'show %s' % config
        if params:
            command += ' %s' % params

        result = self.run_commands(command, 'text')
        if as_string:
            return str(result[0]['output']).strip()

        return str(result[0]['output']).split('\n')

    def refresh(self):
        """Refreshes the instance config properties

        This method will refresh the public running_config and startup_config
        properites.  Since the properties are lazily loaded, this method will
        clear the current internal instance variables.  One the next call the
        instance variables will be repopulated with the current config

        """
        self._running_config = None
        self._startup_config = None

    def configure_session(self):
        """Enter a config session
        """
        self._session_name = self._session_name or uuid4()

    def diff(self):
        """Returns session-config diffs in text encoding

        Note: "show session-config diffs" doesn't support json encoding
        """
        response = self._configure_session(
            ['show session-config diffs'], encoding='text' )

        return response[0]['output']

    def commit(self):
        """Commits the current config session
        """
        return self._configure_and_exit_session(['commit'])

    def abort(self):
        """Aborts the current config session
        """
        return self._configure_session(['abort'])

    def _configure_and_exit_session(self, commands, **kwargs):
        response = self._configure_session(commands, **kwargs)

        if self.autorefresh:
            self.refresh()

        # Exit the current config session
        self._session_name = None

        return response


def connect_to(name):
    """Creates a node instance based on an entry from the config

    This function will retrieve the settings for the specified connection
    from the config and return a Node instance.  The configuration must
    be loaded prior to calling this function.

    Args:
        name (str): The name of the connection to load from the config.  The
            name argument should be the connection name (everything right of
            the colon from the INI file)

    Returns:
        This function will return an instance of Node with the settings
            from the config instance.

    Raises:
        AttributeError: raised if the specified configuration name is not
            found in the loaded configuration

    """
    kwargs = config_for(name)

    if not kwargs:
        raise AttributeError('connection profile not found in config')

    node = connect(return_node=True, **kwargs)
    return node



#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with EOS VRRP resources

The Vrrp resource provides configuration management of interface specific
vrrp resources on and EOS node. It provides the following class
implementations:

    * Vrrp - Configure vrrps in EOS

Vrrp Attributes:
    - enable (boolean): The shutdown state of the vrrp
    - primary_ip (string): The ip address of the vrrp
    - secondary_ip (dict): The secondary ip addresses configured for the vrrp
        This is a dictionary in the format::

            { key: [ list of ip addresses ] }

        where key is 'add', 'remove', or 'exists'. 'add' is
        used to add the list of secondary ip addresses
        to the vrrp. 'remove' will remove the list of
        secondary ip addresses from the vrrp. 'exists' is
        a report only key for retrieving the current
        secondary ip addresses on a vrrp.

    - priority (int): The priority rank of the vrrp
    - description (string): The description for the vrrp
    - ip_version (int): The ip version value for the vrrp
    - timers_advertise (int): The timers advertise setting for the vrrp
    - mac_addr_adv_interval (int): The mac-address advertisement-interval
        setting for the vrrp

    - preempt (boolean): The preempt state of the vrrp
    - preempt_delay_min (int): The preempt delay minimum setting for the vrrp
    - preempt_delay_reload (int): The preempt delay reload setting for the vrrp
    - delay_reload (int): The delay reload setting for the vrrp
    - track (list): The object tracking settings for the vrrp
    - bfd_ip (string): The bfd ip set for the vrrp

Notes:
    The get method will return a dictionary of all the currently configured
    vrrps on a single interface, with the VRID of each vrrp as the keys
    in the dictionary::

        {
            vrrp1: { data },
            vrrp2: { data },
        }

    The getall method will return a dictionary of all the currently configured
    vrrps on the node, with the interface name as the top-level keys, with
    the VRIDs for each vrrp on an interface as a sub-key of that interface::

        {
            interface1: {
                vrrp1: { data },
                vrrp2: { data },
            },
            interface2: {
                vrrp1: { data },
                vrrp2: { data },
            }
        }

    The data for a configured vrrp is a dictionary with the following format::

        {
            enable: <True|False>
            primary_ip: <string>
            priority: <int>
            description: <string|None>
            secondary_ip: {
                exists: [ <ip string1>, <ip string2> ]
                }
            ip_version: <int>
            timers_advertise: <int>
            mac_addr_adv_interval: <int>
            preempt: <True|False>
            preempt_delay_min: <int>
            preempt_delay_reload: <int>
            delay_reload: <int>
            track: [
                {
                    name: <string>
                    action: <shutdown|decrement>
                    amount: <int>|default|no|None
                },
                {
                    name: <string>
                    action: <shutdown|decrement>
                    amount: <int>|default|no|None
                },
            ]
            bfd_ip: <string>
        }

    The create and method accepts a kwargs dictionary which
    defines the properties to be applied to the new or existing vrrp
    configuration. The available keywords and values are as follows:

        - enable: True to enable (no shutdown)|False to disable (shutdown)
        - primary_ip: <ip_string>|no|default|None
        - priority: <int>|no|default|None
        - description: <string>|no|default|None
        - secondary_ip: <dict> may include the following
            - add: <list of ip address strings>
            - remove: <list of ip address strings>
        - ip_version: <int>|no|default|None
        - timers_advertise: <int>|no|default|None
        - mac_addr_adv_interval: <int>|no|default|None
        - preempt: True to enable (preempt)|False to disable (no preempt)
        - preempt_delay_min: <int>|no|default|None
        - preempt_delay_reload: <int>|no|default|None
        - delay_reload: <int>|no|default|None
        - track: <list> of dicts in the following format::

            {
                name: <string>
                action: <shutdown|decrement>
                amount: <int>|default|no|None
            }

        - bfd_ip: <ip string>|no|default|None

"""

import re

from pyeapi.api import EntityCollection

PROPERTIES = ['primary_ip', 'priority', 'description', 'secondary_ip',
              'ip_version', 'enable', 'timers_advertise',
              'mac_addr_adv_interval', 'preempt',
              'preempt_delay_min', 'preempt_delay_reload',
              'delay_reload', 'track', 'bfd_ip']


class Vrrp(EntityCollection):
    """The Vrrp class provides management of the VRRP configuration

    The Vrrp class is derived from EntityCollection and provides an API for
    working with the node's vrrp configurations.
    """

    def get(self, name):
        """Get the vrrp configurations for a single node interface

        Args:
            name (string): The name of the interface for which vrrp
                configurations will be retrieved.

        Returns:
            A dictionary containing the vrrp configurations on the interface.
            Returns None if no vrrp configurations are defined or
            if the interface is not configured.
        """

        # Validate the interface and vrid are specified
        interface = name
        if not interface:
            raise ValueError("Vrrp.get(): interface must contain a value.")

        # Get the config for the interface. Return None if the
        # interface is not defined
        config = self.get_block('interface %s' % interface)
        if config is None:
            return config

        # Find all occurrences of vrids in this interface and make
        # a set of the unique vrid numbers
        match = set(re.findall(r'^\s+(?:no |)vrrp (\d+)', config, re.M))
        if not match:
            return None

        # Initialize the result dict
        result = dict()

        for vrid in match:
            subd = dict()

            # Parse the vrrp configuration for the vrid(s) in the list
            subd.update(self._parse_delay_reload(config, vrid))
            subd.update(self._parse_description(config, vrid))
            subd.update(self._parse_enable(config, vrid))
            subd.update(self._parse_ip_version(config, vrid))
            subd.update(self._parse_mac_addr_adv_interval(config, vrid))
            subd.update(self._parse_preempt(config, vrid))
            subd.update(self._parse_preempt_delay_min(config, vrid))
            subd.update(self._parse_preempt_delay_reload(config, vrid))
            subd.update(self._parse_primary_ip(config, vrid))
            subd.update(self._parse_priority(config, vrid))
            subd.update(self._parse_secondary_ip(config, vrid))
            subd.update(self._parse_timers_advertise(config, vrid))
            subd.update(self._parse_track(config, vrid))
            subd.update(self._parse_bfd_ip(config, vrid))

            result.update({int(vrid): subd})

        # If result dict is empty, return None, otherwise return result
        return result if result else None

    def getall(self):
        """Get the vrrp configurations for all interfaces on a node

        Returns:
            A dictionary containing the vrrp configurations on the node,
            keyed by interface.
        """

        vrrps = dict()

        # Find the available interfaces
        interfaces = re.findall(r'^interface\s(\S+)', self.config, re.M)

        # Get the vrrps defined for each interface
        for interface in interfaces:
            vrrp = self.get(interface)
            # Only add those interfaces that have vrrps defined
            if vrrp:
                vrrps.update({interface: vrrp})

        return vrrps

    def _parse_enable(self, config, vrid):
        if self.version_number >= '4.21.3':
            match = re.search(r'^\s+vrrp %s disabled$' % vrid, config, re.M)
        else:
            match = re.search(r'^\s+vrrp %s shutdown$' % vrid, config, re.M)
        if match:
            return dict(enable=False)
        return dict(enable=True)

    def _parse_primary_ip(self, config, vrid):
        if self.version_number >= '4.21.3':
            match = re.search(r'^\s+vrrp %s ipv4 (\d+\.\d+\.\d+\.\d+)$' %
                              vrid, config, re.M)
        else:
            match = re.search(r'^\s+vrrp %s ip (\d+\.\d+\.\d+\.\d+)$' %
                              vrid, config, re.M)
        value = match.group(1) if match else None
        return dict(primary_ip=value)

    def _parse_priority(self, config, vrid):
        if self.version_number >= '4.21.3':
            match = re.search(r'^\s+vrrp %s priority-level (\d+)$' %
                              vrid, config, re.M)
        else:
            match = re.search(r'^\s+vrrp %s priority (\d+)$' %
                              vrid, config, re.M)
        value = int(match.group(1)) if match else None
        return dict(priority=value)

    def _parse_timers_advertise(self, config, vrid):
        if self.version_number >= '4.21.3':
            match = re.search(r'^\s+vrrp %s advertisement interval (\d+)$' %
                              vrid, config, re.M)
        else:
            match = re.search(r'^\s+vrrp %s timers advertise (\d+)$' %
                              vrid, config, re.M)
        value = int(match.group(1)) if match else None
        return dict(timers_advertise=value)

    def _parse_preempt(self, config, vrid):
        match = re.search(r'^\s+vrrp %s preempt$' % vrid, config, re.M)
        if match:
            return dict(preempt=True)
        return dict(preempt=False)

    def _parse_secondary_ip(self, config, vrid):
        if self.version_number >= '4.21.3':
            matches = re.findall(r'^\s+vrrp %s ipv4 (\d+\.\d+\.\d+\.\d+) '
                                 r'secondary$' % vrid, config, re.M)
        else:
            matches = re.findall(r'^\s+vrrp %s ip (\d+\.\d+\.\d+\.\d+) '
                                 r'secondary$' % vrid, config, re.M)
        value = matches if matches else []
        return dict(secondary_ip=value)

    def _parse_description(self, config, vrid):
        if self.version_number >= '4.21.3':
            match = re.search(r'^\s+vrrp %s session description(.*)$' %
                              vrid, config, re.M)
        else:
            match = re.search(r'^\s+vrrp %s description(.*)$' %
                              vrid, config, re.M)
        if match:
            return dict(description=match.group(1).lstrip())
        return dict(description='')

    def _parse_mac_addr_adv_interval(self, config, vrid):
        match = re.search(r'^\s+vrrp %s mac-address advertisement-interval '
                          r'(\d+)$' % vrid, config, re.M)
        value = int(match.group(1)) if match else None
        return dict(mac_addr_adv_interval=value)

    def _parse_preempt_delay_min(self, config, vrid):
        match = re.search(r'^\s+vrrp %s preempt delay minimum (\d+)$' %
                          vrid, config, re.M)
        value = int(match.group(1)) if match else None
        return dict(preempt_delay_min=value)

    def _parse_preempt_delay_reload(self, config, vrid):
        match = re.search(r'^\s+vrrp %s preempt delay reload (\d+)$' %
                          vrid, config, re.M)
        value = int(match.group(1)) if match else None
        return dict(preempt_delay_reload=value)

    def _parse_bfd_ip(self, config, vrid):
        match = re.search(r'^\s+vrrp %s bfd ip'
                          r'(?: (\d+\.\d+\.\d+\.\d+)|)$' %
                          vrid, config, re.M)
        if match:
            return dict(bfd_ip=match.group(1))
        return dict(bfd_ip='')

    def _parse_ip_version(self, config, vrid):
        if self.version_number >= '4.21.3':
            match = re.search(r'^\s+vrrp %s ipv4 version (\d+)$' %
                              vrid, config, re.M)
        else:
            match = re.search(r'^\s+vrrp %s ip version (\d+)$' %
                              vrid, config, re.M)
        value = int(match.group(1)) if match else None
        return dict(ip_version=value)

    def _parse_delay_reload(self, config, vrid):
        if self.version_number >= '4.21.3':
            match = re.search(r'^\s+vrrp %s timers delay reload (\d+)$' %
                              vrid, config, re.M)
        else:
            match = re.search(r'^\s+vrrp %s delay reload (\d+)$' %
                              vrid, config, re.M)
        value = int(match.group(1)) if match else None
        return dict(delay_reload=value)

    def _parse_track(self, config, vrid):
        if self.version_number >= '4.21.3':
            matches = re.findall(r'^\s+vrrp %s tracked-object (\S+) '
                                 r'(decrement|shutdown)(?:( \d+$|$))' %
                                 vrid, config, re.M)
        else:
            matches = re.findall(r'^\s+vrrp %s track (\S+) '
                                 r'(decrement|shutdown)(?:( \d+$|$))' %
                                 vrid, config, re.M)
        value = []
        for match in matches:
            tr_obj = match[0]
            action = match[1]
            amount = None if match[2] == '' else int(match[2])
            entry = {
                'name': tr_obj,
                'action': action,
            }
            if amount:
                entry.update({'amount': amount})
            value.append(entry)

        # Return the list, sorted for easier comparison
        track_list = sorted(value, key=lambda k: (k['name'], k['action']))
        return dict(track=track_list)

    def create(self, interface, vrid, **kwargs):
        """Creates a vrrp instance from an interface

        Note:
            This method will attempt to create a vrrp in the node's
            operational config. If the vrrp already exists on the
            interface, then this method will set the properties of
            the existing vrrp to those that have been passed in, if
            possible.

        Args:
            interface (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be created.
            kwargs (dict): A dictionary specifying the properties to
                be applied to the new vrrp instance. See library
                documentation for available keys and values.

        Returns:
            True if the vrrp could be created otherwise False (see Node)

        """

        if 'enable' not in kwargs:
            kwargs['enable'] = False

        return self._vrrp_set(interface, vrid, **kwargs)

    def delete(self, interface, vrid):
        """Deletes a vrrp instance from an interface

        Note:
            This method will attempt to delete the vrrp from the node's
            operational config. If the vrrp does not exist on the
            interface then this method will not perform any changes
            but still return True

        Args:
            interface (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be deleted.

        Returns:
            True if the vrrp could be deleted otherwise False (see Node)

        """

        vrrp_str = "no vrrp %d" % vrid
        return self.configure_interface(interface, vrrp_str)

    def default(self, interface, vrid):
        """Defaults a vrrp instance from an interface

        Note:
            This method will attempt to default the vrrp on the node's
            operational config. Default results in the deletion of the
            specified vrrp . If the vrrp does not exist on the
            interface then this method will not perform any changes
            but still return True

        Args:
            interface (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be defaulted.

        Returns:
            True if the vrrp could be defaulted otherwise False (see Node)

        """

        vrrp_str = "default vrrp %d" % vrid
        return self.configure_interface(interface, vrrp_str)

    def set_enable(self, name, vrid, value=False, run=True):
        """Set the enable property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (boolean): True to enable the vrrp, False to disable.
            run (boolean): True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if value is False:
            if self.version_number >= '4.21.3':
                cmd = "vrrp %d disabled" % vrid
            else:
                cmd = "vrrp %d shutdown" % vrid
        elif value is True:
            if self.version_number >= '4.21.3':
                cmd = "no vrrp %d disabled" % vrid
            else:
                cmd = "no vrrp %d shutdown" % vrid
        else:
            raise ValueError("vrrp property 'enable' must be "
                             "True or False")

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_primary_ip(self, name, vrid, value=None, disable=False,
                       default=False, run=True):
        """Set the primary_ip property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (string): IP address to be set.
            disable (boolean): Unset primary ip if True.
            default (boolean): Set primary ip to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if default is True:
            vrrps = self.get(name)
            primary_ip = vrrps[vrid]['primary_ip']
            if self.version_number >= '4.21.3':
                cmd = "default vrrp %d ipv4 %s" % (vrid, primary_ip)
            else:
                cmd = "default vrrp %d ip %s" % (vrid, primary_ip)
        elif disable is True or value is None:
            vrrps = self.get(name)
            primary_ip = vrrps[vrid]['primary_ip']
            if self.version_number >= '4.21.3':
                cmd = "no vrrp %d ipv4 %s" % (vrid, primary_ip)
            else:
                cmd = "no vrrp %d ip %s" % (vrid, primary_ip)
        elif re.match(r'^\d+\.\d+\.\d+\.\d+$', str(value)):
            if self.version_number >= '4.21.3':
                cmd = "vrrp %d ipv4 %s" % (vrid, value)
            else:
                cmd = "vrrp %d ip %s" % (vrid, value)
        else:
            raise ValueError("vrrp property 'primary_ip' must be "
                             "a properly formatted IP address")

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_priority(self, name, vrid, value=None, disable=False,
                     default=False, run=True):
        """Set the primary_ip property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (integer): Priority to assign to the vrrp.
            disable (boolean): Unset priority if True.
            default (boolean): Set priority to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if not str(value).isdigit() or value < 1 or value > 254:
                raise ValueError("vrrp property 'priority' must be "
                                 "an integer in the range 1-254")
        if self.version_number >= '4.21.3':
            cmd = self.command_builder('vrrp %d priority-level' %
                                       vrid, value=value,
                                       default=default, disable=disable)
        else:
            cmd = self.command_builder('vrrp %d priority' %
                                       vrid, value=value,
                                       default=default, disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_description(self, name, vrid, value=None, disable=False,
                        default=False, run=True):
        """Set the description property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (string): Description to assign to the vrrp.
            disable (boolean): Unset description if True.
            default (boolean): Set description to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """
        if self.version_number >= '4.21.3':
            cmd = self.command_builder('vrrp %d session description' %
                                       vrid, value=value,
                                       default=default, disable=disable)
        else:
            cmd = self.command_builder('vrrp %d description' %
                                       vrid, value=value,
                                       default=default, disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_ip_version(self, name, vrid, value=None, disable=False,
                       default=False, run=True):
        """Set the ip_version property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (integer): IP version to assign to the vrrp.
            disable (boolean): Unset ip_version if True.
            default (boolean): Set ip_version to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if value not in (2, 3):
                raise ValueError("vrrp property 'ip_version' must be 2 or 3")
        if self.version_number >= '4.21.3':
            cmd = self.command_builder('vrrp %d ipv4 version' %
                                       vrid, value=value,
                                       default=default, disable=disable)
        else:
            cmd = self.command_builder('vrrp %d ip version' %
                                       vrid, value=value,
                                       default=default, disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_secondary_ips(self, name, vrid, secondary_ips, run=True):
        """Configure the secondary_ip property of the vrrp

        Notes:
            set_secondary_ips takes a list of secondary ip addresses
            which are to be set on the virtal router. An empty list will
            remove any existing secondary ip addresses from the vrrp.
            A list containing addresses will configure the virtual router
            with only the addresses specified in the list - any existing
            addresses not included in the list will be removed.

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            secondary_ips (list): A list of secondary ip addresses to
                be assigned to the virtual router.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        cmds = []

        # Get the current set of tracks defined for the vrrp
        curr_sec_ips = []
        vrrps = self.get(name)
        if vrrps and vrid in vrrps:
            curr_sec_ips = vrrps[vrid]['secondary_ip']

        # Validate the list of ip addresses
        for sec_ip in secondary_ips:
            if type(sec_ip) is not str or \
                    not re.match(r'^\d+\.\d+\.\d+\.\d+$', sec_ip):
                raise ValueError("vrrp property 'secondary_ip' must be a list "
                                 "of properly formatted ip address strings")

        intersection = list(set(curr_sec_ips) & set(secondary_ips))

        # Delete the intersection from both lists to determine which
        # addresses need to be added or removed from the vrrp
        remove = list(set(curr_sec_ips) - set(intersection))
        add = list(set(secondary_ips) - set(intersection))

        # Build the commands to add and remove the secondary ip addresses
        for sec_ip in remove:
            if self.version_number >= '4.21.3':
                cmds.append("no vrrp %d ipv4 %s secondary" % (vrid, sec_ip))
            else:
                cmds.append("no vrrp %d ip %s secondary" % (vrid, sec_ip))

        for sec_ip in add:
            if self.version_number >= '4.21.3':
                cmds.append("vrrp %d ipv4 %s secondary" % (vrid, sec_ip))
            else:
                cmds.append("vrrp %d ip %s secondary" % (vrid, sec_ip))

        cmds = sorted(cmds)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmds)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmds

    def set_timers_advertise(self, name, vrid, value=None, disable=False,
                             default=False, run=True):
        """Set the ip_version property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (integer): Timers advertise value to assign to the vrrp.
            disable (boolean): Unset timers advertise if True.
            default (boolean): Set timers advertise to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if not int(value) or int(value) < 1 or int(value) > 255:
                raise ValueError("vrrp property 'timers_advertise' must be"
                                 "in the range 1-255")
        if self.version_number >= '4.21.3':
            cmd = self.command_builder('vrrp %d advertisement interval' %
                                       vrid,
                                       value=value, default=default,
                                       disable=disable)
        else:
            cmd = self.command_builder('vrrp %d timers advertise' %
                                       vrid,
                                       value=value, default=default,
                                       disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_mac_addr_adv_interval(self, name, vrid, value=None, disable=False,
                                  default=False, run=True):
        """Set the mac_addr_adv_interval property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (integer): mac-address advertisement-interval value to
                assign to the vrrp.
            disable (boolean): Unset mac-address advertisement-interval
                if True.
            default (boolean): Set mac-address advertisement-interval to
                default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if not int(value) or int(value) < 1 or int(value) > 3600:
                raise ValueError("vrrp property 'mac_addr_adv_interval' must "
                                 "be in the range 1-3600")

        cmd = self.command_builder('vrrp %d mac-address advertisement-interval'
                                   % vrid, value=value, default=default,
                                   disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_preempt(self, name, vrid, value=None, disable=False,
                    default=False, run=True):
        """Set the preempt property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (boolean): True to enable preempt, False to disable
                preempt on the vrrp.
            disable (boolean): Unset preempt if True.
            default (boolean): Set preempt to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if value is not True and value is not False:
                raise ValueError("vrrp property 'preempt' must be True "
                                 "or False")

        cmd = self.command_builder('vrrp %d preempt' % vrid, value=value,
                                   default=default, disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_preempt_delay_min(self, name, vrid, value=None, disable=False,
                              default=False, run=True):
        """Set the preempt_delay_min property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (integer): Preempt delay minimum value to set on the vrrp.
            disable (boolean): Unset preempt delay minimum if True.
            default (boolean): Set preempt delay minimum to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if not int(value) or int(value) < 1 or int(value) > 3600:
                raise ValueError("vrrp property 'preempt_delay_min' must be"
                                 "in the range 0-3600 %r" % value)

        cmd = self.command_builder('vrrp %d preempt delay minimum'
                                   % vrid, value=value, default=default,
                                   disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_preempt_delay_reload(self, name, vrid, value=None, disable=False,
                                 default=False, run=True):
        """Set the preempt_delay_min property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (integer): Preempt delay reload value to set on the vrrp.
            disable (boolean): Unset preempt delay reload if True.
            default (boolean): Set preempt delay reload to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if not int(value) or int(value) < 1 or int(value) > 3600:
                raise ValueError("vrrp property 'preempt_delay_reload' must be"
                                 "in the range 0-3600 %r" % value)

        cmd = self.command_builder('vrrp %d preempt delay reload'
                                   % vrid, value=value, default=default,
                                   disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_delay_reload(self, name, vrid, value=None, disable=False,
                         default=False, run=True):
        """Set the preempt_delay_min property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (integer): Preempt delay reload value to set on the vrrp.
            disable (boolean): Unset preempt delay reload if True.
            default (boolean): Set preempt delay reload to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        if not default and not disable:
            if not int(value) or int(value) < 1 or int(value) > 3600:
                raise ValueError("vrrp property 'delay_reload' must be"
                                 "in the range 0-3600 %r" % value)
        if self.version_number >= '4.21.3':
            cmd = self.command_builder('vrrp %d timers delay reload' %
                                       vrid, value=value,
                                       default=default, disable=disable)
        else:
            cmd = self.command_builder('vrrp %d delay reload' %
                                       vrid, value=value,
                                       default=default, disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def set_tracks(self, name, vrid, tracks, run=True):
        """Configure the track property of the vrrp

        Notes:
            set_tracks takes a list of tracked objects which are
            to be set on the virtual router. An empty list will remove
            any existing tracked objects from the vrrp. A list containing
            track entries configures the virtual router to track only the
            objects specified in the list - any existing tracked objects
            on the vrrp not included in the list will be removed.

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            tracks (list): A list of track definition dictionaries. Each
                dictionary is a definition of a tracked object in one
                of the two formats::

                    {'name': tracked_object_name,
                     'action': 'shutdown'}
                    {'name': tracked_object_name,
                     'action': 'decrement',
                     'amount': amount_of_decrement}

            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """

        cmds = []

        # Get the current set of tracks defined for the vrrp
        curr_tracks = []
        vrrps = self.get(name)
        if vrrps and vrid in vrrps:
            curr_tracks = vrrps[vrid]['track']

        # Determine which tracked objects are in both lists using
        # sets of temporary strings built from the track specifications
        unset = '_none_'
        tracks_set = []
        for track in tracks:
            keys = track.keys()

            # Validate no extraneous keys in track definition
            err_keys = set(keys).difference(('name', 'action', 'amount'))
            if err_keys:
                err_keys = ', '.join(err_keys)
                raise ValueError("Error found in vrrp property 'track': "
                                 "unknown key(s) '%s' found. Valid keys are "
                                 "name, action, and amount" % err_keys)

            # Validate required keys in track definition
            if not set(keys).issuperset(('name', 'action')):
                raise ValueError("Error found in vrrp property 'track': "
                                 "track definition must contain 'name' and "
                                 "'action' keys")

            tr_obj = track['name']
            action = track['action']
            amount = track['amount'] if 'amount' in track else unset

            # Validate values in track definition
            error = False
            if action not in ('shutdown', 'decrement'):
                error = True
            if action == 'shutdown' and amount != unset:
                error = True
            if amount != unset and not str(amount).isdigit():
                error = True
            if error:
                raise ValueError("Error found in vrrp property 'track'. "
                                 "See documentation for format specification.")

            tid = "%s   %s   %s" % (tr_obj, action, amount)
            tracks_set.append(tid)

        curr_set = []
        for track in curr_tracks:
            tr_obj = track['name']
            action = track['action']
            amount = track['amount'] if 'amount' in track else unset

            # Validate track definition
            error = False
            if action not in ('shutdown', 'decrement'):
                error = True
            if action == 'shutdown' and amount != unset:
                error = True
            if amount != unset and not str(amount).isdigit():
                error = True
            if error:
                raise ValueError("Error found in vrrp property 'track'. "
                                 "See documentation for format specification.")

            tid = "%s   %s   %s" % (tr_obj, action, amount)
            curr_set.append(tid)

        intersection = list(set(tracks_set) & set(curr_set))

        # Delete the intersection from both lists to determine which
        # track definitions need to be added or removed from the vrrp
        remove = list(set(curr_set) - set(intersection))
        add = list(set(tracks_set) - set(intersection))

        # Build the commands to add and remove the tracked objects
        for track in remove:
            match = re.match(r'(\S+)\s+(\S+)\s+(\S+)', track)
            if match:
                (tr_obj, action, amount) = \
                    (match.group(1), match.group(2), match.group(3))

                if amount == unset:
                    amount = ''
                if self.version_number >= '4.21.3':
                    t_cmd = ("no vrrp %d tracked-object %s %s %s"
                             % (vrid, tr_obj, action, amount))
                else:
                    t_cmd = ("no vrrp %d track %s %s %s"
                             % (vrid, tr_obj, action, amount))
                cmds.append(t_cmd.rstrip())

        for track in add:
            match = re.match(r'(\S+)\s+(\S+)\s+(\S+)', track)
            if match:
                (tr_obj, action, amount) = \
                    (match.group(1), match.group(2), match.group(3))

                if amount == unset:
                    amount = ''
                if self.version_number >= '4.21.3':
                    t_cmd = ("vrrp %d tracked-object %s %s %s"
                             % (vrid, tr_obj, action, amount))
                else:
                    t_cmd = ("vrrp %d track %s %s %s"
                             % (vrid, tr_obj, action, amount))
                cmds.append(t_cmd.rstrip())

        cmds = sorted(cmds)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmds)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmds

    def set_bfd_ip(self, name, vrid, value=None, disable=False,
                   default=False, run=True):
        """Set the bfd_ip property of the vrrp

        Args:
            name (string): The interface to configure.
            vrid (integer): The vrid number for the vrrp to be managed.
            value (string): The bfd ip address to be set.
            disable (boolean): Unset bfd ip if True.
            default (boolean): Set bfd ip to default if True.
            run (boolean): Set to True to execute the command, False to
                return a string with the formatted command.

        Returns:
            If run is True, returns True if the command executed successfully,
            error if failure.

            If run is False, returns the formatted command string which can
            be passed to the node

        """
        if not default and not disable:
            if not re.match(r'^\d+\.\d+\.\d+\.\d+$', str(value)):
                raise ValueError("vrrp property 'bfd_ip' must be "
                                 "a properly formatted IP address")
        cmd = self.command_builder('vrrp %d bfd ip' % vrid, value=value,
                                   default=default, disable=disable)

        # Run the command if requested
        if run:
            result = self.configure_interface(name, cmd)
            # And verify the command succeeded
            if result is False:
                return self.error
            return result

        # Otherwise return the formatted command
        return cmd

    def _vrrp_set(self, name, vrid, **kwargs):
        # Configure the commands to create or update a vrrp
        # configuration, and send the commands to the node.

        vrconf = kwargs

        # step through the individual vrrp properties and
        # set those that need to be changed
        commands = []

        enable = vrconf.get('enable', '__NONE__')
        if enable != '__NONE__':
            cmd = self.set_enable(name, vrid, value=enable, run=False)
            commands.append(cmd)

        primary_ip = vrconf.get('primary_ip', '__NONE__')
        if primary_ip != '__NONE__':
            if primary_ip in ('no', None):
                cmd = self.set_primary_ip(name, vrid, value=None,
                                          disable=True, run=False)
            elif primary_ip == 'default':
                cmd = self.set_primary_ip(name, vrid, value=None,
                                          default=True, run=False)
            else:
                cmd = self.set_primary_ip(name, vrid, value=primary_ip,
                                          run=False)
            commands.append(cmd)

        priority = vrconf.get('priority', '__NONE__')
        if priority != '__NONE__':
            if priority in ('no', None):
                cmd = self.set_priority(name, vrid, value=priority,
                                        disable=True, run=False)
            elif priority == 'default':
                cmd = self.set_priority(name, vrid, value=priority,
                                        default=True, run=False)
            else:
                cmd = self.set_priority(name, vrid, value=priority, run=False)
            commands.append(cmd)

        description = vrconf.get('description', '__NONE__')
        if description != '__NONE__':
            if description in ('no', None):
                cmd = self.set_description(name, vrid, value=description,
                                           disable=True, run=False)
            elif description == 'default':
                cmd = self.set_description(name, vrid, value=description,
                                           default=True, run=False)
            else:
                cmd = self.set_description(name, vrid, value=description,
                                           run=False)
            commands.append(cmd)

        ip_version = vrconf.get('ip_version', '__NONE__')
        if ip_version != '__NONE__':
            if ip_version in ('no', None):
                cmd = self.set_ip_version(name, vrid, value=ip_version,
                                          disable=True, run=False)
            elif ip_version == 'default':
                cmd = self.set_ip_version(name, vrid, value=ip_version,
                                          default=True, run=False)
            else:
                cmd = self.set_ip_version(name, vrid, value=ip_version,
                                          run=False)
            commands.append(cmd)

        secondary_ip = vrconf.get('secondary_ip', '__NONE__')
        if secondary_ip != '__NONE__':
            cmds = self.set_secondary_ips(name, vrid, secondary_ip, run=False)
            for cmd in cmds:
                commands.append(cmd)

        timers_advertise = vrconf.get('timers_advertise', '__NONE__')
        if timers_advertise != '__NONE__':
            if timers_advertise in ('no', None):
                cmd = self.set_timers_advertise(name, vrid,
                                                value=timers_advertise,
                                                disable=True, run=False)
            elif timers_advertise == 'default':
                cmd = self.set_timers_advertise(name, vrid,
                                                value=timers_advertise,
                                                default=True, run=False)
            else:
                cmd = self.set_timers_advertise(name, vrid,
                                                value=timers_advertise,
                                                run=False)
            commands.append(cmd)

        mac_addr_adv_interval = \
            vrconf.get('mac_addr_adv_interval', '__NONE__')
        if mac_addr_adv_interval != '__NONE__':
            if mac_addr_adv_interval in ('no', None):
                cmd = \
                    self.set_mac_addr_adv_interval(name, vrid,
                                                   value=mac_addr_adv_interval,
                                                   disable=True, run=False)
            elif mac_addr_adv_interval == 'default':
                cmd = \
                    self.set_mac_addr_adv_interval(name, vrid,
                                                   value=mac_addr_adv_interval,
                                                   default=True, run=False)
            else:
                cmd = \
                    self.set_mac_addr_adv_interval(name, vrid,
                                                   value=mac_addr_adv_interval,
                                                   run=False)
            commands.append(cmd)

        preempt = vrconf.get('preempt', '__NONE__')
        if preempt != '__NONE__':
            if preempt in ('no', False):
                cmd = self.set_preempt(name, vrid, value=preempt,
                                       disable=True, run=False)
            elif preempt == 'default':
                cmd = self.set_preempt(name, vrid, value=preempt,
                                       default=True, run=False)
            else:
                cmd = self.set_preempt(name, vrid, value=preempt, run=False)
            commands.append(cmd)

        preempt_delay_min = vrconf.get('preempt_delay_min', '__NONE__')
        if preempt_delay_min != '__NONE__':
            if preempt_delay_min in ('no', None):
                cmd = self.set_preempt_delay_min(name, vrid,
                                                 value=preempt_delay_min,
                                                 disable=True, run=False)
            elif preempt_delay_min == 'default':
                cmd = self.set_preempt_delay_min(name, vrid,
                                                 value=preempt_delay_min,
                                                 default=True, run=False)
            else:
                cmd = self.set_preempt_delay_min(name, vrid,
                                                 value=preempt_delay_min,
                                                 run=False)
            commands.append(cmd)

        preempt_delay_reload = vrconf.get('preempt_delay_reload', '__NONE__')
        if preempt_delay_reload != '__NONE__':
            if preempt_delay_reload in ('no', None):
                cmd = self.set_preempt_delay_reload(name, vrid,
                                                    value=preempt_delay_reload,
                                                    disable=True, run=False)
            elif preempt_delay_reload == 'default':
                cmd = self.set_preempt_delay_reload(name, vrid,
                                                    value=preempt_delay_reload,
                                                    default=True, run=False)
            else:
                cmd = self.set_preempt_delay_reload(name, vrid,
                                                    value=preempt_delay_reload,
                                                    run=False)
            commands.append(cmd)

        delay_reload = vrconf.get('delay_reload', '__NONE__')
        if delay_reload != '__NONE__':
            if delay_reload in ('no', None):
                cmd = self.set_delay_reload(name, vrid, value=delay_reload,
                                            disable=True, run=False)
            elif delay_reload == 'default':
                cmd = self.set_delay_reload(name, vrid, value=delay_reload,
                                            default=True, run=False)
            else:
                cmd = self.set_delay_reload(name, vrid, value=delay_reload,
                                            run=False)
            commands.append(cmd)

        track = vrconf.get('track', '__NONE__')
        if track != '__NONE__':
            cmds = self.set_tracks(name, vrid, track, run=False)
            for cmd in cmds:
                commands.append(cmd)

        bfd_ip = vrconf.get('bfd_ip', '__NONE__')
        if bfd_ip != '__NONE__':
            if bfd_ip in ('no', None):
                cmd = self.set_bfd_ip(name, vrid, value=bfd_ip,
                                      disable=True, run=False)
            elif bfd_ip == 'default':
                cmd = self.set_bfd_ip(name, vrid, value=bfd_ip,
                                      default=True, run=False)
            else:
                cmd = self.set_bfd_ip(name, vrid, value=bfd_ip, run=False)
            commands.append(cmd)

        # Send the commands to the requested interface
        result = self.configure_interface(name, commands)
        # And verify the commands succeeded
        if result is False:
            return self.error
        return result

    def vrconf_format(self, vrconfig):
        """Formats a vrrp configuration dictionary to match the
        information as presented from the get and getall methods.
        vrrp configuration dictionaries passed to the create
        method may contain data for setting properties which results
        in a default value on the node. In these instances, the data
        for setting or changing the property is replaced with the
        value that would be returned from the get and getall methods.

        Intended for validating updated vrrp configurations.
        """

        fixed = dict(vrconfig)

        # primary_ip: default, no, None results in address of 0.0.0.0
        if fixed['primary_ip'] in ('no', 'default', None):
            fixed['primary_ip'] = '0.0.0.0'
        # priority: default, no, None results in priority of 100
        if fixed['priority'] in ('no', 'default', None):
            fixed['priority'] = 100
        # description: default, no, None results in None
        if fixed['description'] in ('no', 'default', None):
            fixed['description'] = None
        # secondary_ip: list should be exactly what is required,
        # just sort it for easier comparison
        if 'secondary_ip' in fixed:
            fixed['secondary_ip'] = sorted(fixed['secondary_ip'])
        # ip_version: default, no, None results in value of 2
        if fixed['ip_version'] in ('no', 'default', None):
            fixed['ip_version'] = 2
        # timers_advertise: default, no, None results in value of 1
        if fixed['timers_advertise'] in ('no', 'default', None):
            fixed['timers_advertise'] = 1
        # mac_address_advertisement_interaval:
        #    default, no, None results in value of 30
        if fixed['mac_addr_adv_interval'] in \
                ('no', 'default', None):
            fixed['mac_addr_adv_interval'] = 30
        # preempt: default, no results in value of False
        if fixed['preempt'] in ('no', 'default'):
            fixed['preempt'] = False
        # preempt_delay_min: default, no, None results in value of 0
        if fixed['preempt_delay_min'] in ('no', 'default', None):
            fixed['preempt_delay_min'] = 0
        # preempt_delay_reload: default, no, None results in value of 0
        if fixed['preempt_delay_reload'] in ('no', 'default', None):
            fixed['preempt_delay_reload'] = 0
        # delay_reload: default, no, None results in value of 0
        if fixed['delay_reload'] in ('no', 'default', None):
            fixed['delay_reload'] = 0
        # track: list should be exactly what is required,
        # just sort it for easier comparison
        if 'track' in fixed:
            fixed['track'] = \
                sorted(fixed['track'], key=lambda k: (k['name'], k['action']))
        # bfd_ip: default, no, None results in ''
        if fixed['bfd_ip'] in ('no', 'default', None):
            fixed['bfd_ip'] = ''

        return fixed


def instance(node):
    """Returns an instance of Vrrp

    Args:
        node (Node): The node argument passes an instance of Node to the
            resource

    Returns:
        object: An instance of Vrrp
    """
    return Vrrp(node)



#
# Copyright (c) 2017, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with interfaces in EOS

This module provides an API for pragmatically working with EOS interface
configurations.  Interfaces include any data or management plane interface
available in EOS.

Parameters:
    name (string): The name of the interface the configuration should be
        applied to.  The interface name is the full interface identifier.

    shutdown (boolean): True if the interface is administratively disabled,
        and False if the interface is administratively enable.  This value
        does not validate the interfaces operational state.

    description (string): The interface description string.  This value is
        an arbitrary operator defined value.

    sflow (boolean): True if sFlow is enabled on the interface otherwise
        False

    flowcontrol_send (string): The flowcontrol send configuration value for
        the interface.  Valid values are on or off

    flowcontrol_receive (string): The flowcontrol receive configuration value
        for the interface.  Valid values are on or off

"""

import re

from pyeapi.api import EntityCollection
from pyeapi.utils import ProxyCall, CliVariants

MIN_LINKS_RE = re.compile(r'(?<=\s{3}min-links\s)(?P<value>.+)$', re.M)

DEFAULT_LACP_MODE = 'on'
DEFAULT_LACP_FALLBACK = 'disabled'
DEFAULT_LACP_FALLBACK_TIMEOUT = 90

VALID_INTERFACES = frozenset([
    'Ethernet',
    'Management',
    'Loopback',
    'Port-Channel',
    'Vlan',
    'Vxlan',
])


def isvalidinterface(value):
    match = re.match(r'([EPVLM][a-z-C]+)', value)
    return match and match.group() in VALID_INTERFACES


class Interfaces(EntityCollection):

    def __init__(self, node, *args, **kwargs):
        super(Interfaces, self).__init__(node, *args, **kwargs)
        self._instances = dict()

    def get(self, name):
        return self.get_instance(name)[name]

    def getall(self):
        """Returns all interfaces in a dict object.

        Returns:
            A Python dictionary object containing all interface
            configuration indexed by interface name::

                {
                    "Ethernet1": {...},
                    "Ethernet2": {...}
                }

        """
        interfaces_re = re.compile(r'(?<=^interface\s)(.+)$', re.M)

        response = dict()
        for name in interfaces_re.findall(self.config):
            interface = self.get(name)
            if interface:
                response[name] = interface
        return response

    def __getattr__(self, name):
        return ProxyCall(self.marshall, name)

    def get_instance(self, interface):
        cls = INTERFACE_CLASS_MAP.get(interface[0:2]) or BaseInterface
        if cls in self._instances:
            instance = self._instances[cls]
        else:
            instance = cls(self.node)
            self._instances[cls] = instance
        return instance

    def marshall(self, name, *args, **kwargs):
        interface = args[0]
        if not isvalidinterface(interface):
            raise ValueError('invalid interface {}'.format(interface))

        instance = self.get_instance(interface)
        if not hasattr(instance, name):
            raise AttributeError("'%s' object has no attribute '%s'" %
                                 (instance, name))
        method = getattr(instance, name)
        return method(*args, **kwargs)


class BaseInterface(EntityCollection):

    def __str__(self):
        return 'Interface'

    def get(self, name):
        """Returns a generic interface as a set of key/value pairs

        This class is should normally serve as a  base class for building more
        specific interface resources.  The attributes of this resource are
        common to all interfaces regardless of type in EOS.

        The generic interface resource returns the following:

            * name (str): The name of the interface
            * type (str): Always returns 'generic'
            * shutdown (bool): True if the interface is shutdown
            * description (str): The interface description value

        Args:
            name (str): The interface identifier to retrieve from the
                running-configuration

        Returns:
            A Python dictionary object of key/value pairs that represents
                the interface configuration.  If the specified interface
                does not exist, then None is returned
        """
        config = self.get_block('^interface %s' % name)
        if not config:
            return None

        resource = dict(name=name, type='generic')
        resource.update(self._parse_shutdown(config))
        resource.update(self._parse_description(config))
        return resource

    def _parse_shutdown(self, config):
        """Scans the specified config block and returns the shutdown value

        Args:
            config (str): The interface config block to scan

        Returns:
            dict: Returns a dict object with the shutdown value retrieved
                from the config block.  The returned dict object is intended
                to be merged into the interface resource dict
        """
        value = 'no shutdown' not in config
        return dict(shutdown=value)

    def _parse_description(self, config):
        """Scans the specified config block and returns the description value

        Args:
            config (str): The interface config block to scan

        Returns:
            dict: Returns a dict object with the description value retrieved
                from the config block.  If the description value is not
                configured, None is returned as the value.  The returned dict
                is intended to be merged into the interface resource dict.
        """
        value = None
        match = re.search(r'description (.+)$', config, re.M)
        if match:
            value = match.group(1)
        return dict(description=value)

    def create(self, name):
        """Creates a new interface on the node

        Note:
            This method will attempt to create the interface regardless
            if the interface exists or not.  If the interface already exists
            then this method will still return True

        Args:
            name (string): The full name of the interface.

        Returns:
            True if the interface could be created otherwise False (see Note)

        """
        return self.configure('interface %s' % name)

    def delete(self, name):
        """Deletes the interface from the node

        Note:
            This method will attempt to delete the interface from the nodes
            operational config.  If the interface does not exist then this
            method will not perform any changes but still return True

        Args:
            name (string): The full name of the interface

        Returns:
            True if the interface could be deleted otherwise False (see Node)

        """
        return self.configure('no interface %s' % name)

    def default(self, name):
        """Defaults an interface in the running configuration

        Args:
            name (string): The full name of the interface

        Returns:
            True if the command operation succeeds otherwise False
        """
        return self.configure('default interface %s' % name)

    def set_encapsulation(self, name, vid, default=False, disable=False):
        """Configures the subinterface encapsulation value

        Args:
            name (string): The interface identifier.  It must be a full
                interface name (ie Ethernet, not Et)
            vid (int): The vlan id number
            default (boolean): Specifies to default the subinterface
                encapsulation
            disable (boolean): Specifies to disable the subinterface
                encapsulation

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        if '.' not in name:
            raise NotImplementedError('parameter encapsulation can only be'
                                      ' set on subinterfaces')
        if name[0:2] not in ['Et', 'Po']:
            raise NotImplementedError('parameter encapsulation can only be'
                                      ' set on Ethernet and Port-Channel'
                                      ' subinterfaces')
        commands = ['interface %s' % name]
        commands.append(self.command_builder('encapsulation dot1q vlan',
                                             str(vid), default=default,
                                             disable=disable))
        return self.configure(commands)

    def set_description(self, name, value=None, default=False, disable=False):
        """Configures the interface description

        EosVersion:
            4.13.7M

        Args:
            name (string): The interface identifier.  It must be a full
                interface name (ie Ethernet, not Et)
            value (string): The value to set the description to.
            default (boolean): Specifies to default the interface description
            disable (boolean): Specifies to negate the interface description

        Returns:
            True if the operation succeeds otherwise False
        """
        string = 'description'
        commands = self.command_builder(string, value=value, default=default,
                                        disable=disable)
        return self.configure_interface(name, commands)

    def set_shutdown(self, name, default=False, disable=True):
        """Configures the interface shutdown state

        Default configuration for set_shutdown is disable=True, meaning
        'no shutdown'. Setting both default and disable to False will
        effectively enable shutdown on the interface.

        Args:
            name (string): The interface identifier.  It must be a full
                interface name (ie Ethernet, not Et)

            default (boolean): Specifies to default the interface shutdown

            disable (boolean): Specifies to disable interface shutdown, i.e.
                disable=True => no shutdown

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        commands = ['interface %s' % name]
        commands.append(self.command_builder('shutdown', value=True,
                                             default=default, disable=disable))
        return self.configure(commands)


class EthernetInterface(BaseInterface):

    def __str__(self):
        return 'EthernetInterface'

    def get(self, name):
        """Returns an interface as a set of key/value pairs

        Args:
            name (string): the interface identifier to retrieve the from
                the configuration

        Returns:
            A Python dictionary object of key/value pairs that represent
            the current configuration for the specified node.  If the
            specified interface name does not exist, then None is returned::

                {
                    "name": <string>,
                    "type": "ethernet",
                    "sflow": [true, false],
                    "flowcontrol_send": [on, off],
                    "flowcontrol_receive": [on, off]
                }
        """
        config = self.get_block('^interface %s' % name)

        if not config:
            return None

        resource = super(EthernetInterface, self).get(name)
        resource.update(dict(name=name, type='ethernet'))
        resource.update(self._parse_sflow(config))
        resource.update(self._parse_flowcontrol_send(config))
        resource.update(self._parse_flowcontrol_receive(config))
        return resource

    def _parse_sflow(self, config):
        """Scans the specified config block and returns the sflow value

        Args:
            config (str): The interface config block to scan

        Returns:
            dict: Returns a dict object with the sflow value retrieved
                from the config block.  The returned dict object is intended
                to be merged into the interface resource dict
        """
        value = 'no sflow' not in config
        return dict(sflow=value)

    def _parse_flowcontrol_send(self, config):
        """Scans the config block and returns the flowcontrol send value

        Args:
            config (str): The interface config block to scan

        Returns:
            dict: Returns a dict object with the flowcontrol send value
                retrieved from the config block.  The returned dict object
                is intended to be merged into the interface resource dict
        """
        value = 'off'
        match = re.search(r'flowcontrol send (\w+)$', config, re.M)
        if match:
            value = match.group(1)
        return dict(flowcontrol_send=value)

    def _parse_flowcontrol_receive(self, config):
        """Scans the config block and returns the flowcontrol receive value

        Args:
            config (str): The interface config block to scan

        Returns:
            dict: Returns a dict object with the flowcontrol receive value
                retrieved from the config block.  The returned dict object
                is intended to be merged into the interface resource dict
        """
        value = 'off'
        match = re.search(r'flowcontrol receive (\w+)$', config, re.M)
        if match:
            value = match.group(1)
        return dict(flowcontrol_receive=value)

    def create(self, name):
        """Create an Ethernet subinterface

        Args:
            name (string): The subinterface name. Ex: Ethernet1.1

        Raises:
            NotImplementedError: creating physical Ethernet interfaces is not
            supported. Only subinterfaces can be created.
        """
        if '.' not in name:
            raise NotImplementedError('creating physical Ethernet interfaces'
                                      ' is not supported. Only subinterfaces'
                                      ' can be created')
        return self.configure(['interface %s' % name])

    def delete(self, name):
        """Delete an Ethernet subinterfaces

        Args:
            name (string): The subinterface name. Ex: Ethernet1.1

        Raises:
            NotImplementedError: creating physical Ethernet interfaces is not
            supported. Only subinterfaces can be created.
        """
        if '.' not in name:
            raise NotImplementedError('deleting physical Ethernet interfaces'
                                      ' is not supported. Only subinterfaces'
                                      ' can be created')
        return self.configure(['no interface %s' % name])

    def set_flowcontrol_send(self, name, value=None, default=False,
                             disable=False):
        """Configures the interface flowcontrol send value

        Args:
            name (string): The interface identifier.  It must be a full
                interface name (ie Ethernet, not Et)

            value (boolean): True if the interface should enable sending flow
                control packets, otherwise False

            default (boolean): Specifies to default the interface flow
                control send value

            disable (boolean): Specifies to disable the interface flow
                control send value

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        return self.set_flowcontrol(name, 'send', value, default, disable)

    def set_flowcontrol_receive(self, name, value=None, default=False,
                                disable=False):
        """Configures the interface flowcontrol receive value

        Args:
            name (string): The interface identifier.  It must be a full
                interface name (ie Ethernet, not Et)

            value (boolean): True if the interface should enable receiving
                flow control packets, otherwise False

            default (boolean): Specifies to default the interface flow
                control receive value

            disable (boolean): Specifies to disable the interface flow
                control receive value

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        return self.set_flowcontrol(name, 'receive', value, default, disable)

    def set_flowcontrol(self, name, direction, value=None, default=False,
                        disable=False):
        """Configures the interface flowcontrol value

        Args:
            name (string): The interface identifier.  It must be a full
                interface name (ie Ethernet, not Et)

            direction (string): one of either 'send' or 'receive'

            value (boolean): True if the interface should enable flow control
                packet handling, otherwise False

            default (boolean): Specifies to default the interface flow control
                send or receive value

            disable (boolean): Specifies to disable the interface flow control
                send or receive value

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        if value is not None:
            if value not in ['on', 'off']:
                raise ValueError('invalid flowcontrol value')

        if direction not in ['send', 'receive']:
            raise ValueError('invalid direction specified')

        commands = ['interface %s' % name]
        commands.append(self.command_builder('flowcontrol %s' % direction,
                                             value=value, default=default,
                                             disable=disable))
        return self.configure(commands)

    def set_sflow(self, name, value=None, default=False, disable=False):
        """Configures the sFlow state on the interface

        Args:
            name (string): The interface identifier.  It must be a full
                interface name (ie Ethernet, not Et)

            value (boolean): True if sFlow should be enabled otherwise False

            default (boolean): Specifies the default value for sFlow

            disable (boolean): Specifies to disable sFlow

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        if value not in [True, False, None]:
            raise ValueError

        commands = ['interface %s' % name]
        commands.append(self.command_builder('sflow enable', value=value,
                                             default=default, disable=disable))
        return self.configure(commands)

    def set_vrf(self, name, vrf, default=False, disable=False):
        """Applies a VRF to the interface

           Note: VRF being applied to interface must already exist in switch
               config. Ethernet port must be in routed mode. This functionality
               can also be handled in the VRF api.

           Args:
               name (str): The interface identifier.  It must be a full
                   interface name (ie Ethernet, not Et)
               vrf (str): The vrf name to be applied to the interface
               default (bool): Specifies the default value for VRF
               disable (bool): Specifies to disable VRF

           Returns:
               True if the operation succeeds otherwise False is returned
        """
        commands = ['interface %s' % name]
        if self.version_number >= '4.23':
            commands.append(self.command_builder('vrf', vrf,
                                                 default=default,
                                                 disable=disable))
        else:
            commands.append(self.command_builder('vrf forwarding', vrf,
                                                 default=default,
                                                 disable=disable))
        return self.configure(commands)

class PortchannelInterface(BaseInterface):

    def __str__(self):
        return 'PortchannelInterface'

    def get(self, name):
        """Returns a Port-Channel interface as a set of key/value pairs

        Args:
            name (str): The interface identifier to retrieve from the
                running-configuration

        Returns:
            A Python dictionary object of key/value pairs that represents
            the interface configuration.  If the specified interface
            does not exist, then None is returned::

                {
                    "name": <string>,
                    "type": "portchannel",
                    "members": <arrary of interface names>,
                    "minimum_links: <integer>,
                    "lacp_mode": [on, active, passive]
                }

        """
        config = self.get_block('^interface %s' % name)
        if not config:
            return None

        response = super(PortchannelInterface, self).get(name)
        response.update(dict(name=name, type='portchannel'))

        response['members'] = self.get_members(name)
        response['lacp_mode'] = self.get_lacp_mode(name)
        response.update(self._parse_minimum_links(config))
        response.update(self._parse_lacp_timeout(config))
        response.update(self._parse_lacp_fallback(config))
        return response

    def _parse_minimum_links(self, config):
        value = 0
        match = re.search(r'port-channel min-links (\d+)', config)
        if match:
            value = int(match.group(1))
        return dict(minimum_links=value)

    def _parse_lacp_fallback(self, config):
        value = DEFAULT_LACP_FALLBACK
        match = re.search(r'lacp fallback (static|individual)', config)
        if match:
            value = match.group(1)
        return dict(lacp_fallback=value)

    def _parse_lacp_timeout(self, config):
        value = DEFAULT_LACP_FALLBACK_TIMEOUT
        match = re.search(r'lacp fallback timeout (\d+)', config)
        if match:
            value = int(match.group(1))
        return dict(lacp_timeout=value)

    def get_lacp_mode(self, name):
        """Returns the LACP mode for the specified Port-Channel interface

        Args:
            name(str): The Port-Channel interface name to return the LACP
                mode for from the configuration

        Returns:
            The configured LACP mode for the interface.  Valid mode values
                are 'on', 'passive', 'active'

        """
        members = self.get_members(name)
        if not members:
            return DEFAULT_LACP_MODE

        for member in self.get_members(name):
            match = re.search(r'channel-group\s\d+\smode\s(?P<value>.+)',
                              self.get_block('^interface %s' % member))
            return match.group('value')

    def get_members(self, name):
        """Returns the member interfaces for the specified Port-Channel

        Args:
            name(str): The Port-channel interface name to return the member
                interfaces for

        Returns:
            A list of physical interface names that belong to the specified
                interface
        """
        grpid = re.search(r'(\d+)', name).group()
        command = 'show port-channel %s all-ports' % grpid
        config = self.node.enable(command, 'text')
        return re.findall(r'\b(?!Peer)Ethernet[\d/]*\b',
                          config[0]['result']['output'])

    def set_members(self, name, members, mode=None):
        """Configures the array of member interfaces for the Port-Channel

        Args:
            name(str): The Port-Channel interface name to configure the member
                interfaces

            members(list): The list of Ethernet interfaces that should be
                member interfaces

            mode(str): The LACP mode to configure the member interfaces to.
                Valid values are 'on, 'passive', 'active'. When there are
                existing channel-group members and their lacp mode differs
                from this attribute, all of those members will be removed and
                then re-added using the specified lacp mode. If this attribute
                is omitted, the existing lacp mode will be used for new
                member additions.

        Returns:
            True if the operation succeeds otherwise False
        """
        commands = list()
        grpid = re.search(r'(\d+)', name).group()
        current_members = self.get_members(name)
        lacp_mode = self.get_lacp_mode(name)
        if mode and mode != lacp_mode:
            lacp_mode = mode
            self.set_lacp_mode(grpid, lacp_mode)

        # remove members from the current port-channel interface
        for member in set(current_members).difference(members):
            commands.append('interface %s' % member)
            commands.append('no channel-group %s' % grpid)

        # add new member interfaces to the port-channel interface
        for member in set(members).difference(current_members):
            commands.append('interface %s' % member)
            commands.append('channel-group %s mode %s' % (grpid, lacp_mode))

        return self.configure(commands) if commands else True

    def set_lacp_mode(self, name, mode):
        """Configures the LACP mode of the member interfaces

        Args:
            name(str): The Port-Channel interface name to configure the
                LACP mode

            mode(str): The LACP mode to configure the member interfaces to.
                Valid values are 'on, 'passive', 'active'

        Returns:
            True if the operation succeeds otherwise False
        """
        if mode not in ['on', 'passive', 'active']:
            return False

        grpid = re.search(r'(\d+)', name).group()

        remove_commands = list()
        add_commands = list()

        for member in self.get_members(name):
            remove_commands.append('interface %s' % member)
            remove_commands.append('no channel-group %s' % grpid)
            add_commands.append('interface %s' % member)
            add_commands.append('channel-group %s mode %s' % (grpid, mode))

        return self.configure(remove_commands + add_commands)

    def set_minimum_links(self, name, value=None, default=False,
                          disable=False):
        """Configures the Port-Channel min-links value

        Args:
            name(str): The Port-Channel interface name

            value(str): The value to configure the min-links

            default (bool): Specifies to default the port channel min-links
                value

            disable (bool): Specifies to disable the port channel min-links
                value

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        commands = ['interface %s' % name]
        commands.append(self.command_builder('port-channel min-links',
                                             value=value, default=default,
                                             disable=disable))
        return self.configure(commands)

    def set_lacp_fallback(self, name, mode=None):
        """Configures the Port-Channel lacp_fallback

        Args:
            name(str): The Port-Channel interface name

            mode(str): The Port-Channel LACP fallback setting
                Valid values are 'disabled', 'static', 'individual':

                * static  - Fallback to static LAG mode
                * individual - Fallback to individual ports
                * disabled - Disable LACP fallback

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        if mode not in ['disabled', 'static', 'individual']:
            return False
        disable = True if mode == 'disabled' else False
        commands = ['interface %s' % name]
        commands.append(self.command_builder('port-channel lacp fallback',
                                             value=mode, disable=disable))
        return self.configure(commands)

    def set_lacp_timeout(self, name, value=None):
        """Configures the Port-Channel LACP fallback timeout
           The fallback timeout configures the period an interface in
           fallback mode remains in LACP mode without receiving a PDU.

        Args:
            name(str): The Port-Channel interface name

            value(int): port-channel lacp fallback timeout in seconds

        Returns:
            True if the operation succeeds otherwise False is returned
        """
        commands = ['interface %s' % name]
        string = 'port-channel lacp fallback timeout'
        commands.append(self.command_builder(string, value=value))
        return self.configure(commands)


class VxlanInterface(BaseInterface):

    DEFAULT_SRC_INTF = ''
    DEFAULT_MCAST_GRP = ''

    def __str__(self):
        return 'VxlanInterface'

    def get(self, name):
        """Returns a Vxlan interface as a set of key/value pairs

        The Vxlan interface resource returns the following:

            * name (str): The name of the interface
            * type (str): Always returns 'vxlan'
            * source_interface (str): The vxlan source-interface value
            * multicast_group (str): The vxlan multicast-group value
            * udp_port (int): The vxlan udp-port value
            * vlans (dict): The vlan to vni mappings
            * flood_list (list): The list of global VTEP flood list
            * multicast_decap (bool): If the multicast decap
                                      feature is configured

        Args:
            name (str): The interface identifier to retrieve from the
                running-configuration

        Returns:
            A Python dictionary object of key/value pairs that represents
                the interface configuration.  If the specified interface
                does not exist, then None is returned
        """
        config = self.get_block('^interface %s' % name)
        if not config:
            return None

        response = super(VxlanInterface, self).get(name)
        response.update(dict(name=name, type='vxlan'))

        response.update(self._parse_source_interface(config))
        response.update(self._parse_multicast_group(config))
        response.update(self._parse_udp_port(config))
        response.update(self._parse_vlans(config))
        response.update(self._parse_flood_list(config))
        response.update(self._parse_multicast_decap(config))

        return response

    def _parse_source_interface(self, config):
        """ Parses the conf block and returns the vxlan source-interface value

        Parses the provided configuration block and returns the value of
        vxlan source-interface.  If the value is not configured, this method
        will return DEFAULT_SRC_INTF instead.

        Args:
            config (str): The Vxlan config block to scan

        Return:
            dict: A dict object intended to be merged into the resource dict
        """
        match = re.search(r'vxlan source-interface ([^\s]+)', config)
        value = match.group(1) if match else self.DEFAULT_SRC_INTF
        return dict(source_interface=value)

    def _parse_multicast_group(self, config):
        match = re.search(r'vxlan multicast-group '
                          r'([\d]{3}\.[\d]+\.[\d]+\.[\d]+)',
                          config)
        value = match.group(1) if match else self.DEFAULT_MCAST_GRP
        return dict(multicast_group=value)

    def _parse_multicast_decap(self, config):
        val1 = 'vxlan multicast-group decap' in config
        val2 = 'no vxlan multicast-group decap' in config
        return dict( multicast_decap=bool(val1 ^ val2) )

    def _parse_udp_port(self, config):
        match = re.search(r'vxlan udp-port (\d+)', config)
        value = int(match.group(1))
        return dict(udp_port=value)

    def _parse_vlans(self, config):
        vlans = frozenset(re.findall(r'vxlan vlan (\d+)', config))
        values = dict()

        for vid in vlans:
            values[vid] = dict()

            regexp = r'vxlan vlan {} vni (\d+)'.format(vid)
            match = re.search(regexp, config)
            values[vid]['vni'] = match.group(1) if match else None

            regexp = r'vxlan vlan {} flood vtep (.*)$'.format(vid)
            matches = re.search(regexp, config, re.M)
            flood_list = matches.group(1).split(' ') if matches else []
            values[vid]['flood_list'] = flood_list

        return dict(vlans=values)

    def _parse_flood_list(self, config):
        match = re.search(r'^ *vxlan flood vtep +([\d. ]+)$', config, re.M)
        values = list()
        if match:
            values = match.group(1).split(' ')
        return dict(flood_list=values)

    def set_source_interface(self, name, value=None, default=False,
                             disable=False):
        """Configures the Vxlan source-interface value

        EosVersion:
            4.13.7M

        Args:
            name(str): The interface identifier to configure, defaults to
                Vxlan1
           value(str): The value to configure the source-interface to
           default(bool): Configures the source-interface value to default
           disable(bool): Negates the source-interface

        Returns:
            True if the operation succeeds otherwise False
        """
        string = 'vxlan source-interface'
        cmds = self.command_builder(string, value=value, default=default,
                                    disable=disable)
        return self.configure_interface(name, cmds)

    def set_multicast_group(self, name, value=None, default=False,
                            disable=False):
        """Configures the Vxlan multicast-group value

        EosVersion:
            4.13.7M

        Args:
            name(str): The interface identifier to configure, defaults to
                Vxlan1
           value(str): The value to configure the multicast-group to
           default(bool): Configures the mulitcast-group value to default
           disable(bool): Negates the multicast-group value

        Returns:
            True if the operation succeeds otherwise False
        """
        string_dpr = 'vxlan multicast-group'
        cmds_dpr = self.command_builder(string_dpr,
            value=value, default=default, disable=disable)
        string_new = 'vxlan multicast-group decap'
        cmds_new = self.command_builder(string_new,
            value=value, default=default, disable=disable)
        return self.configure_interface(name,
            CliVariants(cmds_new, cmds_dpr) )

    def set_multicast_decap(self, name, default=False,
                            disable=False):
        """Configures the Vxlan multicast-group decap feature

        EosVersion:
            4.15.0M

        Args:
            name(str): The interface identifier to configure, defaults to
                Vxlan1
           default(bool): Configures the mulitcast-group decap value to default
           disable(bool): Negates the multicast-group decap value

        Returns:
            True if the operation succeeds otherwise False
        """
        string = 'vxlan multicast-group decap'
        if default or disable:
            cmds = self.command_builder(string, value=None, default=default,
                                        disable=disable)
        else:
            cmds = [string]
        return self.configure_interface(name, cmds)


    def set_udp_port(self, name, value=None, default=False, disable=False):
        """Configures vxlan udp-port value

        EosVersion:
            4.13.7M

        Args:
            name(str): The name of the interface to configure
            value(str): The value to set udp-port to
            default(bool): Configure using the default keyword
            disable(bool): Negate the udp-port value

        Returns:
            True if the operation succeeds otherwise False
        """
        string = 'vxlan udp-port'
        cmds = self.command_builder(string, value=value, default=default,
                                    disable=disable)
        return self.configure_interface(name, cmds)

    def add_vtep(self, name, vtep, vlan=None):
        """Adds a new VTEP endpoint to the global or local flood list

        EosVersion:
            4.13.7M

        Args:
            name (str): The name of the interface to configure
            vtep (str): The IP address of the remote VTEP endpoint to add
            vlan (str): The VLAN ID associated with this VTEP.  If the VLAN
            keyword is used, then the VTEP is configured as a local flood
            endpoing

        Returns:
            True if the command completes successfully
        """
        if not vlan:
            cmd = 'vxlan flood vtep add {}'.format(vtep)
        else:
            cmd = 'vxlan vlan {} flood vtep add {}'.format(vlan, vtep)
        return self.configure_interface(name, cmd)

    def remove_vtep(self, name, vtep, vlan=None):
        """Removes a VTEP endpoint from the global or local flood list

        EosVersion:
            4.13.7M

        Args:
            name (str): The name of the interface to configure
            vtep (str): The IP address of the remote VTEP endpoint to add
            vlan (str): The VLAN ID associated with this VTEP.  If the VLAN
            keyword is used, then the VTEP is configured as a local flood
            endpoing

        Returns:
            True if the command completes successfully
        """
        if not vlan:
            cmd = 'vxlan flood vtep remove {}'.format(vtep)
        else:
            cmd = 'vxlan vlan {} flood vtep remove {}'.format(vlan, vtep)
        return self.configure_interface(name, cmd)

    def update_vlan(self, name, vid, vni):
        """Adds a new vlan to vni mapping for the interface

        EosVersion:
            4.13.7M

        Args:
            vlan (str, int): The vlan id to map to the vni
            vni (str, int): The vni value to use

        Returns:
            True if the command completes successfully

        """
        cmd = 'vxlan vlan add %s vni %s' % (vid, vni)
        return self.configure_interface(name, cmd)

    def remove_vlan(self, name, vid):
        """Removes a vlan to vni mapping for the interface

        EosVersion:
            4.13.7M

        Args:
            vlan (str, int): The vlan id to map to the vni

        Returns:
            True if the command completes successfully

        """
        return self.configure_interface( name,
            CliVariants(f'vxlan vlan remove {vid} vni $',
                f'vxlan vlan remove {vid} vni') )


INTERFACE_CLASS_MAP = {
    'Et': EthernetInterface,
    'Po': PortchannelInterface,
    'Vx': VxlanInterface
}


def instance(api):
    return Interfaces(api)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working with logical layer 2 switchports in EOS

This module provides an API for working with logical layer 2 interfaces
(switchports) in EOS.  Switchports are interfaces built on top of
physical Ethernet and bundled Port-Channel interfaces.

"""

import re

from pyeapi.api import EntityCollection
from pyeapi.utils import make_iterable


class Switchports(EntityCollection):
    """The Switchports class provides a configuration resource for swichports

    Logical layer 2 interfaces built on top of physical Ethernet and bundled
    Port-Channel interfaces can be configured and managed with an instance
    of Switchports.   The Switchports class is a resource collection and
    supports get and getall methods.  The Switchports class is derived from
    the BaseResource class

    """

    def get(self, name):
        """Returns a dictionary object that represents a switchport

        The Switchport resource returns the following:

            * name (str): The name of the interface
            * mode (str): The switchport mode value
            * access_vlan (str): The switchport access vlan value
            * trunk_native_vlan (str): The switchport trunk native vlan vlaue
            * trunk_allowed_vlans (str): The trunk allowed vlans value
            * trunk_groups (list): The list of trunk groups configured

        Args:
            name (string): The interface identifier to get.  Note: Switchports
                are only supported on Ethernet and Port-Channel interfaces

        Returns:
            dict: A Python dictionary object of key/value pairs that represent
                the switchport configuration for the interface specified  If
                the specified argument is not a switchport then None
                is returned
        """
        config = self.get_block('interface %s' % name)
        if 'no switchport\n' in config:
            return

        resource = dict(name=name)
        resource.update(self._parse_mode(config))
        resource.update(self._parse_access_vlan(config))
        resource.update(self._parse_trunk_native_vlan(config))
        resource.update(self._parse_trunk_allowed_vlans(config))
        resource.update(self._parse_trunk_groups(config))
        return resource

    def _parse_mode(self, config):
        """Scans the specified config and parses the switchport mode value

        Args:
            config (str): The interface configuration block to scan

        Returns:
            dict: A Python dict object with the value of switchport mode.
                The dict returned is intended to be merged into the resource
                dict
        """
        value = re.search(r'switchport mode (\w+)', config, re.M)
        return dict(mode=value.group(1))

    def _parse_trunk_groups(self, config):
        """Scans the specified config and parses the trunk group values

        Args:
            config (str): The interface configuraiton blcok

        Returns:
            A dict object with the trunk group values that can be merged
                into the resource dict
        """
        values = re.findall(r'switchport trunk group ([^\s]+)', config, re.M)
        return dict(trunk_groups=values)

    def _parse_access_vlan(self, config):
        """Scans the specified config and parse the access-vlan value
        Args:
            config (str): The interface configuration block to scan

        Returns:
            dict: A Python dict object with the value of switchport access
                value.  The dict returned is intended to be merged into the
                resource dict
        """
        value = re.search(r'switchport access vlan (\d+)', config)
        return dict(access_vlan=value.group(1) if value else None)

    def _parse_trunk_native_vlan(self, config):
        """Scans the specified config and parse the trunk native vlan value

        Args:
            config (str): The interface configuration block to scan

        Returns:
            dict: A Python dict object with the value of switchport trunk
                native vlan value.  The dict returned is intended to be
                merged into the resource dict
        """
        match = re.search(r'switchport trunk native vlan (\d+)', config)
        return dict(trunk_native_vlan=match.group(1))

    def _parse_trunk_allowed_vlans(self, config):
        """Scans the specified config and parse the trunk allowed vlans value

        Args:
            config (str): The interface configuration block to scan

        Returns:
            dict: A Python dict object with the value of switchport trunk
                allowed vlans value.  The dict returned is intended to be
                merged into the resource dict
        """
        match = re.search(r'switchport trunk allowed vlan (.+)$', config, re.M)
        return dict(trunk_allowed_vlans=match.group(1))

    def getall(self):
        """Returns a dict object to all Switchports

        This method will return all of the configured switchports as a
        dictionary object keyed by the interface identifier.

        Returns:
            A Python dictionary object that represents all configured
                switchports in the current running configuration
        """
        interfaces_re = re.compile(r'(?<=^interface\s)([Et|Po][^.\s]+)$', re.M)

        response = dict()
        for name in interfaces_re.findall(self.config):
            interface = self.get(name)
            if interface:
                response[name] = interface
        return response

    def create(self, name):
        """Creates a new logical layer 2 interface

        This method will create a new switchport for the interface specified
        in the arguments (name).  If the logical switchport already exists
        then this command will have no effect

        Args:
            name (string): The interface identifier to create the logical
                layer 2 switchport for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1)

        Returns:
            True if the create operation succeeds otherwise False.  If the
                interface specified in args is already a switchport then this
                method will have no effect but will still return True
        """
        commands = ['interface %s' % name, 'no ip address',
                    'switchport']
        return self.configure(commands)

    def delete(self, name):
        """Deletes the logical layer 2 interface

        This method will delete the logical switchport for the interface
        specified in the arguments.  If the interface doe not have a logical
        layer 2 interface defined, then this method will have no effect.

        Args:
            name (string): The interface identifier to create the logical
                layer 2 switchport for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1)

        Returns:
            True if the create operation succeeds otherwise False.  If the
                interface specified in args is already a switchport then this
                method will have no effect but will still return True
        """
        commands = ['interface %s' % name, 'no switchport']
        return self.configure(commands)

    def default(self, name):
        """Defaults the configuration of the switchport interface

        This method will default the configuration state of the logical
        layer 2 interface.

        Args:
            name (string): The interface identifier to create the logical
                layer 2 switchport for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1)

        Returns:
            True if the create operation succeeds otherwise False.  If the
                interface specified in args is already a switchport then this
                method will have no effect but will still return True
        """
        commands = ['interface %s' % name, 'no ip address',
                    'default switchport']
        return self.configure(commands)

    def set_mode(self, name, value=None, default=False, disable=False):
        """Configures the switchport mode

        Args:
            name (string): The interface identifier to create the logical
                layer 2 switchport for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1)

            value (string): The value to set the mode to.  Accepted values
                for this argument are access or trunk

            default (bool): Configures the mode parameter to its default
                value using the EOS CLI

            disable (bool): Negate the mode parameter using the EOS CLI

        Returns:
            True if the create operation succeeds otherwise False.
        """
        string = 'switchport mode'
        command = self.command_builder(string, value=value, default=default,
                                       disable=disable)
        return self.configure_interface(name, command)

    def set_access_vlan(self, name, value=None, default=False, disable=False):
        """Configures the switchport access vlan

        Args:
            name (string): The interface identifier to create the logical
                layer 2 switchport for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1)

            value (string): The value to set the access vlan to.  The value
                must be a valid VLAN ID in the range of 1 to 4094.

            default (bool): Configures the access vlan parameter to its default
                value using the EOS CLI

            disable (bool): Negate the access vlan parameter using the EOS CLI

        Returns:
            True if the create operation succeeds otherwise False.
        """
        string = 'switchport access vlan'
        command = self.command_builder(string, value=value, default=default,
                                       disable=disable)
        return self.configure_interface(name, command)

    def set_trunk_native_vlan(self, name, value=None, default=False,
                              disable=False):
        """Configures the switchport trunk native vlan value

        Args:
            name (string): The interface identifier to create the logical
                layer 2 switchport for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1)

            value (string): The value to set the trunk nativevlan to.  The
                value must be a valid VLAN ID in the range of 1 to 4094.

            default (bool): Configures the access vlan parameter to its default
                value using the EOS CLI

            disable (bool): Negate the access vlan parameter using the EOS CLI

        Returns:
            True if the create operation succeeds otherwise False.
        """
        string = 'switchport trunk native vlan'
        command = self.command_builder(string, value=value, default=default,
                                       disable=disable)
        return self.configure_interface(name, command)

    def set_trunk_allowed_vlans(self, name, value=None, default=False,
                                disable=False):
        """Configures the switchport trunk allowed vlans value

        Args:
            name (string): The interface identifier to create the logical
                layer 2 switchport for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1)

            value (string): The value to set the trunk allowed vlans to.  The
                value must be a valid VLAN ID in the range of 1 to 4094.

            default (bool): Configures the access vlan parameter to its default
                value using the EOS CLI

            disable (bool): Negate the access vlan parameter using the EOS CLI

        Returns:
            True if the create operation succeeds otherwise False.
        """
        string = 'switchport trunk allowed vlan'
        command = self.command_builder(string, value=value, default=default,
                                       disable=disable)
        return self.configure_interface(name, command)

    def set_trunk_groups(self, intf, value=None, default=False, disable=False):
        """Configures the switchport trunk group value

        Args:
            intf (str): The interface identifier to configure.
            value (str): The set of values to configure the trunk group
            default (bool): Configures the trunk group default value
            disable (bool): Negates all trunk group settings

        Returns:
            True if the config operation succeeds otherwise False
        """
        if default:
            cmd = 'default switchport trunk group'
            return self.configure_interface(intf, cmd)

        if disable:
            cmd = 'no switchport trunk group'
            return self.configure_interface(intf, cmd)

        current_value = self.get(intf)['trunk_groups']
        failure = False

        value = make_iterable(value)

        for name in set(value).difference(current_value):
            if not self.add_trunk_group(intf, name):
                failure = True

        for name in set(current_value).difference(value):
            if not self.remove_trunk_group(intf, name):
                failure = True

        return not failure

    def add_trunk_group(self, intf, value):
        """Adds the specified trunk group to the interface

        Args:
            intf (str): The interface name to apply the trunk group to
            value (str): The trunk group value to apply to the interface

        Returns:
            True if the operation as successfully applied otherwise false
        """
        string = 'switchport trunk group {}'.format(value)
        return self.configure_interface(intf, string)

    def remove_trunk_group(self, intf, value):
        """Removes a specified trunk group to the interface

        Args:
            intf (str): The interface name to remove the trunk group from
            value (str): The trunk group value

        Returns:
            True if the operation as successfully applied otherwise false
        """
        string = 'no switchport trunk group {}'.format(value)
        return self.configure_interface(intf, string)


def instance(node):
    """Returns an instance of Switchports

    This method will create and return an instance of the Switchports object
    passing the value of node to the instance.  The module method is
    required for the resource to be autoloaded by the Node object

    Args:
        node (Node): The node argument provides an instance of Node to the
            resource
    """
    return Switchports(node)



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Module for working the logical IP interfaces in EOS

This module provides an API for configuring logical IP interfaces using
EOS and eAPI.

Parameters:
    name (string): The interface name the configuration is in reference
        to.  The interface name is the full interface identifier.

    address (string): The interface IP address in the form of
        address/len.

    mtu (integer): The interface MTU value.  The MTU value accepts
        integers in the range of 68 to 65535 bytes.  See RFC 791 and
        RFC 8200 for more information.
"""

import re

from pyeapi.api import EntityCollection
from pyeapi.utils import _interpolate_docstr

IP_MTU_MIN = 68
IP_MTU_MAX = 65535

SWITCHPORT_RE = re.compile(r'no switchport$', re.M)


class Ipinterfaces( EntityCollection ):

    def get( self, name ):
        """Returns the specific IP interface properties

        The Ipinterface resource returns the following:

            * name (str): The name of the interface
            * address (str): The IP address of the interface in the form
                of A.B.C.D/E (None if no ip configured)
            * secondary (list): The list of secondary IP addresses of the
                interface (if any configured)
            * mtu (int): The configured value for IP MTU.


        Args:
            name (string): The interface identifier to retrieve the
                configuration for

        Return:
            A Python dictionary object of key/value pairs that represents
                the current configuration of the node.  If the specified
                interface does not exist then None is returned.
        """
        config = self.get_block( 'interface %s' % name )
        if name[ 0:2 ] in [
                'Et', 'Po' ] and not SWITCHPORT_RE.search( config, re.M ):
            return None

        resource = dict( name=name )
        resource.update( self._parse_address(config) )
        resource.update( self._parse_mtu(config) )
        return resource

    def _parse_address( self, config ):
        """Parses the config block and returns the ip address value

        The provided configuration block is scanned and the configured value
        for the IP address is returned as a dict object. If the IP address
        value is not configured, then None is returned for the value

        Args:
            config (str): The interface configuration block to parse

        Return:
            dict: A dict object intended to be merged into the resource dict
        """
        match = re.findall( r'ip address ([^\s]+)', config, re.M )
        primary, secondary = ( match[0],
                match[1:] ) if match else ( None, None )
        return dict( address=primary,
         secondary=secondary ) if secondary else dict( address=primary )

    def _parse_mtu(self, config):
        """Parses the config block and returns the configured IP MTU value

        The provided configuration block is scanned and the configured value
        for the IP MTU is returned as a dict object.  The IP MTU value is
        expected to always be present in the provided config block

        Args:
            config (str): The interface configuration block to parse

        Return:
            dict: A dict object intended to be merged into the resource dict
        """
        match = re.search(r'mtu (\d+)', config)
        return dict( mtu=int(match.group( 1 )) if match else None )

    def getall(self):
        """ Returns all of the IP interfaces found in the running-config

        Returns:
            A Python dictionary object of key/value pairs keyed by interface
            name that represents all of the IP interfaces on
            the current node::

                {
                    'Ethernet1': {...},
                    'Ethernet2': {...}
                }
        """
        interfaces_re = re.compile(r'^interface\s(.+)', re.M)

        response = dict()
        for name in interfaces_re.findall(self.config):
            interface = self.get(name)
            if interface:
                response[name] = interface
        return response

    def create(self, name):
        """ Creates a new IP interface instance

        This method will create a new logical IP interface for the specified
        physical interface.   If a logical IP interface already exists then
        this operation will have no effect.

        Note:
            Configuring a logical IP interface on a physical interface will
            remove any existing logical switchports have have been created

        Args:
            name (string): The interface identifier to create the logical
                layer 3 IP interface for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1).

        Returns:
            True if the create operation succeeds otherwise False.  If the
                specified interface is already created the this method will
                have no effect but will still return True
        """
        commands = ['interface %s' % name, 'no switchport']
        return self.configure(commands)

    def delete(self, name):
        """ Deletes an IP interface instance from the running configuration

        This method will delete the logical IP interface for the specified
        physical interface.  If the interface does not have a logical
        IP interface defined, then this method will have no effect.

        Args:
            name (string): The interface identifier to create the logical
                layer 3 IP interface for.  The name must be the full interface
                name and not an abbreviated interface name (eg Ethernet1, not
                Et1).

        Returns:
            True if the delete operation succeeds otherwise False.

        """
        commands = ['interface %s' % name, 'no ip address', 'switchport']
        return self.configure(commands)

    def set_address(self, name, value=None, default=False, disable=False):
        """ Configures the interface IP address

        Args:
            name (string): The interface identifier to apply the interface
                config to

            value (string): The IP address and mask to set the interface to.
                The value should be in the format of A.B.C.D/E

            default (bool): Configures the address parameter to its default
                value using the EOS CLI default command

            disable (bool): Negates the address parameter value using the
                EOS CLI no command

        Returns:
            True if the operation succeeds otherwise False.

        """
        commands = ['interface %s' % name]
        commands.append(self.command_builder('ip address', value=value,
                                             default=default, disable=disable))
        return self.configure(commands)

    @_interpolate_docstr( 'IP_MTU_MIN', 'IP_MTU_MAX' )
    def set_mtu(self, name, value=None, default=False, disable=False):
        """ Configures the interface IP MTU

        Args:
            name (string): The interface identifier to apply the interface
                config to

            value (integer): The MTU value to set the interface to.  Accepted
                values include IP_MTU_MIN to IP_MTU_MAX

            default (bool): Configures the mtu parameter to its default
                value using the EOS CLI default command

            disable (bool); Negate the mtu parameter value using the EOS
                CLI no command

        Returns:
            True if the operation succeeds otherwise False.

        Raises:
            ValueError: If the value for MTU is not an integer value or
                outside of the allowable range

        """
        if value is not None:
            value = int(value)
            if not IP_MTU_MIN <= value <= IP_MTU_MAX:
                raise ValueError('invalid mtu value')

        commands = ['interface %s' % name]
        commands.append(self.command_builder('mtu', value=value,
                                             default=default, disable=disable))
        return self.configure(commands)


def instance(node):
    """Returns an instance of Ipinterfaces

    This method will create and return an instance of the Ipinterfaces object
    passing the value of node to the instance.  This function is required
    for the resource to be autoloaded by the Node object

    Args:
        node (Node): The node argument provides an instance of Node to
            the Ipinterfaces instance
    """
    return Ipinterfaces(node)
